#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def sobol_index(s=None, ns=None, ya=None, yb=None, yc=None,
                si=True, sti=True,
                mean=False, wmean=False,
                method='Mai1999'):
    """
        Calculates the first-order Si and total STi variance-based sensitivity indices
        summarised in Saltelli et al. (2010) with improvements of Mai et al. (2014).


        Definition
        ----------
        def sobol_index(s=None, ns=None, ya=None, yb=None, yc=None,
                        si=True, sti=True,
                        mean=False, wmean=False,
                        method='Mai1999'):


        Optional Input
        --------------
        Either
          s        ns*(k+2) model outputs
          ns       base sample number
        or
          ya       ns model outputs f(A)
          yb       second ns model outputs f(B)
          yc       (ns,k) model outputs f(A_B) or f(B_A)
        with k: # or parameters. See Saltelli et al. (2008, 2010) for definitions.
        yc can be either f(B_A), i.e. yb with one column of ya
        or vice versa f(A_B), i.e. ya with one column of yb. See Saltelli et al. (2010).

        If 2D input (3D for yc), then Si/STi will be calculated for each element in first dimension.
        Assumes that first dimension is time, i.e. calculates indices per time step.
        mean and wmean give then the mean and variance weighted mean of the time steps, resp.

        si         if True (default): output Si
        sti        if True (default): output STi
        mean       if True: output mean Si and/or STi (if 2D/3D input)
        wmean      if True: output variance weighted mean Si and/or STi (if 2D/3D input)
        method     string, case-insensitive (default: 'Mai1999')
                   'Saltelli2008' - The formulation presented in 'The Primer'. (yc=f(B_A))
                                    Si  = (1/n*sum_j(f(A)_j*f(B_A^i)_j) - mean(f(A))^2)/var(f(A))
                                    STi = (var(f(A))-(1/n*sum_j(f(B)_j*f(B_A^i)_j) - mean(f(A))^2))/var(f(A))
                   'Homma1996'    - Si takes mixed A, B term for squared mean in nominator. (yc=f(B_A))
                                    Si  = 1/n*sum_j(f(A)_j*(f(B_A^i)_j - f(B)_j))/var(f(A))
                                    STi as Saltelli2008
                   'Sobol2007'    - STi takes A, A_B instead of B, B_A (needs f(A_B) and f(B_A), not avail.)
                                    Si as Homma1996
                                    STi = 1/n*sum_j(f(A)_j*(f(A_B^i)_j - f(A)_j))/var(f(A))
                   'Saltelli2010' - Si takes B, A_B instead of A, B_A (yc=f(A_B))
                                    Si  = 1/n*sum_j(f(B)_j*(f(A_B^i)_j - f(A)_j))/var(f(A))
                                    STi as Sobol2007
                   'Jansen1999'   - Calculate Si and STi by expectation(variance) instead of variance(expectation) (yc=f(A_B))
                                    Si  = (var(f(A)) - 1/2n*sum_j(f(B)_j - f(A_B^i)_j)^2)/var(f(A))
                                    STi = 1/2n*sum_j(f(A)_j - f(A_B^i)_j)^2/var(f(A))
                   'Mai2012'      - Si of Homma 1996 but denominator is variance of A and B (yc=f(B_A)) and
                                    STi of Homma 1996 but mean and variance of f(A) replaced by f(B)
                                    Si  = 1/n*sum_j(f(A)_j*(f(B_A^i)_j - f(B)_j))/var([f(A),f(B)])
                                    STi = (var(f(A))-(1/n*sum_j(f(B)_j*f(B_A^i)_j) - mean(f(B)^2)))/var(f(B))
                   'Mai2013'      - Si of Saltelli2010 but denominator is variance of A and B (yc=f(A_B))
                                    Si  = 1/n*sum_j(f(B)_j*(f(A_B^i)_j - f(A)_j))/var([f(A),f(B)])
                                    STi as Sobol2007/Saltelli2010
                   'Mai2014'      - SI of Jansen1999  but denominator is variance of A and B (yc=f(A_B))
                                    Si  = (var([f(A),f(B)]) - 1/2n*sum_j(f(B)_j - f(A_B^i)_j)^2)/var([f(A),f(B)])
                                    STi as Jansen1999
                   'Mai1999'      - SI of Mai2013 (Saltelli2010 with var([f(A),f(B)])) and STi of Jansen1999 (yc=f(A_B))
                                    Si  = 1/n*sum_j(f(B)_j*(f(A_B^i)_j - f(A)_j))/var([f(A),f(B)])
                                    STi = 1/2n*sum_j(f(A)_j - f(A_B^i)_j)^2/var(f(A))


        Output
        ------
        First-order sensitivity indices Si and Total sensitivity indices STi.


        Restrictions
        ------------
        1. The right order and size is assumed if s and ns are given.
        2. If three positional arguments are given, i.e. would be s, ns and ya, then it is assumed
           that the user provided ya, yb and yc.
        3. yc is either f(A_B) or f(B_A) depending on the method.
        4. method='Sobol2007' would need f(A_B) and f(B_A) and is therefore not available here.


        References
        ----------
        Mai et al. (2014) ...
        Saltelli, A. et al. (2008) Global sensitivity analysis. The primer.
            John Wiley & Sons Inc., NJ, USA, ISBN 978-0-470-05997-5 (pp. 1-292)
        Saltelli, A. et al. (2010), Variance based sensitivity analysis of model output. Design and estimator
            for the total sensitivity index. Computer Physics Communications 181, 259-270.


        Examples
        --------

        >>> import numpy as np
        >>> import scipy.stats as stats
        >>> from lhs import lhs
        >>> from autostring import astr

        >>> # ------------------------------------------------------------------------
        >>> #                                1D
        >>> #
        >>> # seed for reproducible results in doctest
        >>> np.random.seed(1)
        >>> # number of parameter sets
        >>> nsets = 1000
        >>> # number of parameters
        >>> npara = 3
        >>> # randomly sample nsets with Latin-Hypercube
        >>> dist = [stats.uniform, stats.uniform, stats.uniform]
        >>> pars = [(-np.pi,2.0*np.pi),(-np.pi,2.0*np.pi),(-np.pi,2.0*np.pi)]
        >>> setsa  = lhs(dist, pars, nsets)
        >>> setsb  = lhs(dist, pars, nsets)
        >>> # Saltelli has two constants in the model
        >>> a = 0.5
        >>> b = 2.0
        >>> # generate model output yA and yB
        >>> iya = np.sin(setsa[0,:])  + a*(np.sin(setsa[1,:]))**2  + b*setsa[2,:]**4  * np.sin(setsa[0,:])
        >>> iyb = np.sin(setsb[0,:])  + a*(np.sin(setsb[1,:]))**2  + b*setsb[2,:]**4  * np.sin(setsb[0,:])
        >>> # generate model output yCi = f(B_A)
        >>> iyc = np.empty((npara,nsets))*np.nan
        >>> for i in range(npara):
        ...     tmpset = np.copy(setsb)
        ...     tmpset[i,:] = setsa[i,:]
        ...     iyc[i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + b*tmpset[2,:]**4  * np.sin(tmpset[0,:])
        >>> # generate model output yCi = f(A_B)
        >>> iyc2 = np.empty((npara,nsets))*np.nan
        >>> for i in range(npara):
        ...     tmpset = np.copy(setsa)
        ...     tmpset[i,:] = setsb[i,:]
        ...     iyc2[i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + b*tmpset[2,:]**4  * np.sin(tmpset[0,:])

        >>> # Theoretical results: has to converge for n->Infinity to this values
        >>> vy = 0.5 + a**2/8.0 + b*np.pi**4/5.0+b**2*np.pi**8/18.0
        >>> theo_si  = np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0)/vy, (a**2/8.0)/vy, 0.0])
        >>> theo_sti = np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0+8.0/225.0*b**2*np.pi**8)/vy,
        ...                      (a**2/8.0)/vy, (8.0/225.0*b**2*np.pi**8)/vy])
        >>> # Theoretical results
        >>> print('T :: si  =',astr(theo_si,3,pp=True))
        T :: si  = ['0.372' '0.000' '0.000']
        >>> print('T :: sti =',astr(theo_sti,3,pp=True))
        T :: sti = ['1.000' '0.000' '0.628']

        >>> # 3 positional arguments
        >>> isi1, isti1 = sobol_index(iya,iyb,iyc2)
        >>> print('S :: si  =',astr(isi1,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']
        >>> print('S :: sti =',astr(isti1,3,pp=True))
        S :: sti = ['1.014' '0.000' '0.654']

        >>> # 3 optional arguments
        >>> isi2, isti2 = sobol_index(ya=iya,yb=iyb,yc=iyc2)
        >>> print('S :: si  =',astr(isi2,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']

        >>> # 2 positional arguments
        >>> s2 = np.concatenate((iya, iyb, np.ravel(iyc2)))
        >>> ns = iya.size
        >>> isi3, isti3 = sobol_index(s2, ns)
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']

        >>> # 2 optional arguments
        >>> isi4, isti4 = sobol_index(s=s2, ns=ns)
        >>> print('S :: si  =',astr(isi4,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']

        >>> # 2 optional arguments and no STi output
        >>> isi5 = sobol_index(s=s2, ns=ns, si=True, sti=False)
        >>> print('S :: si  =',astr(isi5,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']

        >>> # 2 optional arguments and no Si output
        >>> isti5 = sobol_index(s=s2, ns=ns, si=False, sti=True)
        >>> print('S :: sti =',astr(isti5,3,pp=True))
        S :: sti = ['1.014' '0.000' '0.654']

        >>> # Method=Saltelli2008
        >>> s = np.concatenate((iya, iyb, np.ravel(iyc)))
        >>> isi3, isti3 = sobol_index(s, ns, method='Saltelli2008')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = ['0.367' '0.011' '0.016']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = [' 1.038' '-0.019' ' 0.641']

        >>> # Method=Homma1996
        >>> isi3, isti3 = sobol_index(s, ns, method='homma1996')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.356' ' 0.000' ' 0.005']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = [' 1.038' '-0.019' ' 0.641']

        >>> # Method=Mai2012
        >>> isi3, isti3 = sobol_index(s, ns, method='mai2012')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.353' ' 0.000' ' 0.005']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = ['1.037' '0.000' '0.647']

        >>> # Method=saltelli2010
        >>> isi3, isti3 = sobol_index(s2, ns, method='saltelli2010')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.348' ' 0.000' '-0.049']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = [' 0.984' ' 0.000' ' 0.633']

        >>> # Method=jansen1999
        >>> isi3, isti3 = sobol_index(s2, ns, method='jansen1999')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.319' ' 0.002' '-0.068']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = ['1.014' '0.000' '0.654']

        >>> # Method=mai2013
        >>> isi3, isti3 = sobol_index(s2, ns, method='mai2013')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = [' 0.984' ' 0.000' ' 0.633']

        >>> # Method=mai2014
        >>> isi3, isti3 = sobol_index(s2, ns, method='mai2014')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.325' ' 0.011' '-0.059']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = ['1.014' '0.000' '0.654']

        >>> # Method=mai
        >>> isi3, isti3 = sobol_index(s2, ns, method='mai1999')
        >>> print('S :: si  =',astr(isi3,3,pp=True))
        S :: si  = [' 0.345' ' 0.000' '-0.049']
        >>> print('S :: sti =',astr(isti3,3,pp=True))
        S :: sti = ['1.014' '0.000' '0.654']

        >>> # ------------------------------------------------------------------------
        >>> #                                2D
        >>> #
        >>> # number of timepoints
        >>> ntime = 100
        >>> iya   = np.empty((ntime,nsets))*np.nan
        >>> iyb   = np.empty((ntime,nsets))*np.nan
        >>> iyc   = np.empty((ntime,npara,nsets))*np.nan
        >>> iyc2  = np.empty((ntime,npara,nsets))*np.nan
        >>> s     = np.empty((ntime,(npara+2)*nsets))*np.nan
        >>> s2    = np.empty((ntime,(npara+2)*nsets))*np.nan
        >>> ns    = nsets
        >>> theo_si  = []
        >>> theo_sti = []
        >>> for t in range(ntime):
        ...     a = 1.0*t + 50.0
        ...     b = 2.0
        ...     iya[t,:] = np.sin(setsa[0,:])  + a*(np.sin(setsa[1,:]))**2  + b*setsa[2,:]**4  * np.sin(setsa[0,:])
        ...     iyb[t,:] = np.sin(setsb[0,:])  + a*(np.sin(setsb[1,:]))**2  + b*setsb[2,:]**4  * np.sin(setsb[0,:])
        ...     for i in range(npara):
        ...         tmpset = np.copy(setsb)
        ...         tmpset[i,:] = setsa[i,:]
        ...         iyc[t,i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + (
        ...                      b*tmpset[2,:]**4  * np.sin(tmpset[0,:]))
        ...     for i in range(npara):
        ...         tmpset = np.copy(setsa)
        ...         tmpset[i,:] = setsb[i,:]
        ...         iyc2[t,i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + (
        ...                       b*tmpset[2,:]**4  * np.sin(tmpset[0,:]))
        ...     s[t,:]  = np.concatenate((iya[t,:], iyb[t,:], np.ravel(iyc[t,:,:])))
        ...     s2[t,:] = np.concatenate((iya[t,:], iyb[t,:], np.ravel(iyc2[t,:,:])))
        ...     vy = 0.5 + a**2/8.0 + b*np.pi**4/5.0+b**2*np.pi**8/18.0
        ...     theo_si  += [np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0)/vy, (a**2/8.0)/vy, 0.0])]
        ...     theo_sti += [np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0+8.0/225.0*b**2*np.pi**8)/vy,
        ...                            (a**2/8.0)/vy, (8.0/225.0*b**2*np.pi**8)/vy])]

        >>> # 1D model output --> ya, yb are 2D and yc is 3D
        >>> isi1, isti1 = sobol_index(iya,iyb,iyc2)

        >>> # SI :: 1st time point
        >>> print('S :: si[t0]  =',astr(isi1[0,:],3,pp=True))
        S :: si[t0]  = [' 0.290' ' 0.140' '-0.031']
        >>> print('T :: si[t0]  =',astr(theo_si[0],3,pp=True))
        T :: si[t0]  = ['0.325' '0.127' '0.000']
        >>> # SI :: 3rd time point
        >>> print('S :: si[t2]  =',astr(isi1[2,:],3,pp=True))
        S :: si[t2]  = [' 0.286' ' 0.149' '-0.030']
        >>> print('T :: si[t2]  =',astr(theo_si[2],3,pp=True))
        T :: si[t2]  = ['0.321' '0.136' '0.000']
        >>> # STI :: 1st time point
        >>> print('S :: sti[t0] =',astr(isti1[0,:],3,pp=True))
        S :: sti[t0] = ['0.881' '0.126' '0.568']
        >>> print('T :: sti[t0] =',astr(theo_sti[0],3,pp=True))
        T :: sti[t0] = ['0.873' '0.127' '0.548']
        >>> # STI :: 3rd time point
        >>> print('S :: sti[t2] =',astr(isti1[2,:],3,pp=True))
        S :: sti[t2] = ['0.871' '0.134' '0.562']
        >>> print('T :: sti[t2] =',astr(theo_sti[2],3,pp=True))
        T :: sti[t2] = ['0.864' '0.136' '0.543']

        >>> # Mean
        >>> isi2, isti2, msi2, msti2 = sobol_index(ya=iya,yb=iyb,yc=iyc2, mean=True)
        >>> print('S :: si_m    =',astr(msi2,3,pp=True))
        S :: si_m    = [' 0.207' ' 0.367' '-0.015']
        >>> print('S :: sti_m   =',astr(msti2,3,pp=True))
        S :: sti_m   = ['0.643' '0.351' '0.415']

        >>> # Weighted mean
        >>> isi4, isti4, wsi2, wsti2 = sobol_index(s2, ns, wmean=True)
        >>> print('S :: si_w    =',astr(wsi2,3,pp=True))
        S :: si_w    = [' 0.197' ' 0.394' '-0.014']
        >>> print('S :: sti_w   =',astr(wsti2,3,pp=True))
        S :: sti_w   = ['0.616' '0.377' '0.397']

        >>> # 2 positional arguments
        >>> isi3, isti3 = sobol_index(s=s2, ns=ns)
        >>> # SI :: 3rd time point
        >>> print('S :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        S :: si[t2]  = [' 0.286' ' 0.149' '-0.030']
        >>> print('T :: si[t2]  =',astr(theo_si[2],3,pp=True))
        T :: si[t2]  = ['0.321' '0.136' '0.000']

        >>> # Method=Saltelli2008
        >>> isi3, isti3 = sobol_index(s, ns, method='saltelli2008')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = ['0.354' '0.130' '0.011']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = ['0.352' '0.139' '0.010']

        >>> # Method=Homma1996
        >>> isi3, isti3 = sobol_index(s, ns, method='homma1996')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.342' ' 0.118' '-0.002']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.340' ' 0.127' '-0.002']

        >>> # Method=Mai2012
        >>> isi3, isti3 = sobol_index(s, ns, method='Mai2012')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.336' ' 0.116' '-0.002']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.334' ' 0.125' '-0.002']

        >>> # Method=Saltelli2010
        >>> isi3, isti3 = sobol_index(s2, ns, method='saltelli2010')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.294' ' 0.142' '-0.031']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.291' ' 0.152' '-0.031']

        >>> # Method=Jansen1999
        >>> isi3, isti3 = sobol_index(s2, ns, method='jansen1999')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.284' ' 0.133' '-0.072']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.281' ' 0.143' '-0.072']

        >>> # Method=Mai2013
        >>> isi3, isti3 = sobol_index(s2, ns, method='Mai2013')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.290' ' 0.140' '-0.031']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.286' ' 0.149' '-0.030']

        >>> # Method=Mai2014
        >>> isi3, isti3 = sobol_index(s2, ns, method='mai2014')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.295' ' 0.147' '-0.056']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.293' ' 0.156' '-0.055']

        >>> # Method=Mai
        >>> isi3, isti3 = sobol_index(s2, ns, method='mai1999')
        >>> print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
        Sal :: si[t0]  = [' 0.290' ' 0.140' '-0.031']
        >>> print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
        Sal :: si[t2]  = [' 0.286' ' 0.149' '-0.030']

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2012
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Aug 2013 - Si per time step
                  JM, Sep 2013 - changes of Mai et al. (2014), theoretical example in docstring
                  MC, Sep 2013 - saltelli
                  MC, Sep 2013 - method, removed saltelli
                  MC, Apr 2014 - assert
    """
    # Check input
    assert (si+sti) > 0, 'No output chosen: si=False and sti=False.'
    # s, ns or ya, yb, yc
    isone = False
    if ((s is not None) and (ns is not None) and (ya is None)):
        if s.ndim == 1:
            isone = True
            s = s[np.newaxis,:]
        nsa   = ns
        ntime = s.shape[0]
        ss    = s.shape[1]
        nn    = (ss//nsa)-2
        iyA   = s[:,0:nsa]
        iyB   = s[:,nsa:2*nsa]
        iyC   = np.reshape(s[:,2*nsa:],(ntime,nn,nsa))
    else:
        assert (((s is not None) and (ns is not None) and (ya is not None)) or ((ya is not None) and (yb is not None) and (yc is not None))), 'Either s and ns has to be given or ya, yb and yc.'
        if ((s is not None) and (ns is not None) and (ya is not None)):
            if s.ndim == 1:
                isone = True
                s  = s[np.newaxis,:]
                ns = ns[np.newaxis,:]
                ya = ya[np.newaxis,:,:]
            iyA     = s
            iyB     = ns
            iyC     = ya
        else:
            if ya.ndim == 1:
                isone = True
                ya = ya[np.newaxis,:]
                yb = yb[np.newaxis,:]
                yc = yc[np.newaxis,:,:]
            iyA     = ya
            iyB     = yb
            iyC     = yc
        ntime = iyA.shape[0]
        nsa   = iyA.shape[1]
        if not ((nsa == iyB.shape[1]) & (nsa == iyC.shape[2])):
            raise ValueError('ya and yb must have same size as yc[1].')
        nn = iyC.shape[1]

    mm = method.lower()
    if mm == 'saltelli2008':
        meanA = np.mean(iyA, axis=1)
        varA  = np.var(iyA,  axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = (np.mean(iyA*iyCi, axis=1) - meanA**2) / varA
            isti[:,i] = 1. - (np.mean(iyB*iyCi, axis=1) - meanA**2) / varA
    elif mm == 'homma1996':
        meanA = np.mean(iyA, axis=1)
        varA  = np.var(iyA,  axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = np.mean(iyA*(iyCi - iyB), axis=1) / varA
            isti[:,i] = 1. - (np.mean(iyB*iyCi, axis=1) - meanA**2) / varA
            #old code isti[:,i] = 1. - (np.mean(iyB*(iyCi-iyA), axis=1)) / varA
    elif mm == 'sobol2007':
        raise ValueError('Sobol2007 would need f(A_B) and f(B_A). It is thus not implemented here.')
    elif mm == 'saltelli2010':
        varA  = np.var(iyA,  axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = np.mean(iyB*(iyCi - iyA), axis=1) / varA
            isti[:,i] = np.mean(iyA*(iyA - iyCi), axis=1) / varA
    elif mm == 'jansen1999':
        varA  = np.var(iyA,  axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = 1. - np.sum((iyB -iyCi)**2, axis=1) / (2.*nsa * varA)
            isti[:,i] = np.sum((iyA -iyCi)**2, axis=1) / (2.*nsa * varA)
    elif mm == 'mai2012':
        meanB = np.mean(iyB, axis=1)
        varB  = np.var(iyB,  axis=1)
        varAB = np.var(np.append(iyA, iyB, axis=1), axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = np.mean(iyA*(iyCi - iyB), axis=1) / varAB
            isti[:,i] = 1. - (np.mean(iyB*iyCi, axis=1) - meanB**2) / varB
    elif mm == 'mai2013':
        varA  = np.var(iyA,  axis=1)
        varAB = np.var(np.append(iyA, iyB, axis=1), axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = np.mean(iyB*(iyCi - iyA), axis=1) / varAB
            isti[:,i] = np.mean(iyA*(iyA - iyCi), axis=1) / varA
    elif mm == 'mai2014':
        varA  = np.var(iyA,  axis=1)
        varAB = np.var(np.append(iyA, iyB, axis=1), axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = 1. - np.sum((iyB -iyCi)**2, axis=1) / (2.*nsa * varAB)
            isti[:,i] = np.sum((iyA -iyCi)**2, axis=1) / (2.*nsa * varA)
    elif mm == 'mai1999':
        varA  = np.var(iyA,  axis=1)
        varAB = np.var(np.append(iyA, iyB, axis=1), axis=1)
        isi  = np.empty((ntime,nn))
        isti = np.empty((ntime,nn))
        for i in range(nn):
            iyCi      = iyC[:,i,:]
            isi[:,i]  = np.mean(iyB*(iyCi - iyA), axis=1) / varAB
            isti[:,i] = np.sum((iyA -iyCi)**2, axis=1) / (2.*nsa * varA)
    else:
        raise ValueError('method unknown: {0}.'.format(method))

    if not isone:
        # simple mean
        if mean:
            msi  = np.mean(isi,  axis=0)
            msti = np.mean(isti, axis=0)

        # weighted mean
        if wmean:
            if (mm == 'mai2012'):
                denomAB = 1./np.sum(varAB)
                denomB  = 1./np.sum(varB)
                varB    = varB[:, np.newaxis]
                varAB   = varAB[:,np.newaxis]
                wsi     = np.sum(isi*varAB, axis=0) * denomAB
                wsti    = np.sum(isti*varB, axis=0) * denomB
            elif (mm == 'mai1999') | (mm == 'mai2013') | (mm == 'mai2014'):
                denomAB = 1./np.sum(varAB)
                denomA  = 1./np.sum(varA)
                varA    = varA[:, np.newaxis]
                varAB   = varAB[:,np.newaxis]
                wsi     = np.sum(isi*varAB, axis=0) * denomAB
                wsti    = np.sum(isti*varA, axis=0) * denomA
            else:
                denom = 1./np.sum(varA)
                varA  = varA[:,np.newaxis]
                wsi   = np.sum(isi*varA,  axis=0) * denom
                wsti  = np.sum(isti*varA, axis=0) * denom

    # Output
    if isone:
        isi  = isi[0,:]
        isti = isti[0,:]

    out = []
    if si:  out += [isi]
    if sti: out += [isti]
    if not isone:
        if mean:
            if si:  out += [msi]
            if sti: out += [msti]
        if wmean:
            if si:  out += [wsi]
            if sti: out += [wsti]

    if len(out) == 1:
        return out[0]
    else:
        return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np
    # import scipy.stats as stats
    # from lhs import lhs
    # from autostring import astr

    # # ------------------------------------------------------------------------
    # #                                1D
    # #
    # # seed for reproducible results in doctest
    # np.random.seed(1)
    # # number of parameter sets
    # nsets = 1000
    # # number of parameters
    # npara = 3
    # # randomly sample nsets with Latin-Hypercube
    # dist = [stats.uniform, stats.uniform, stats.uniform]
    # pars = [(-np.pi,2.0*np.pi),(-np.pi,2.0*np.pi),(-np.pi,2.0*np.pi)]
    # setsa  = lhs(dist, pars, nsets)
    # setsb  = lhs(dist, pars, nsets)
    # # Saltelli has two constants in the model
    # a = 0.5
    # b = 2.0
    # # generate model output yA and yB
    # iya = np.sin(setsa[0,:])  + a*(np.sin(setsa[1,:]))**2  + b*setsa[2,:]**4  * np.sin(setsa[0,:])
    # iyb = np.sin(setsb[0,:])  + a*(np.sin(setsb[1,:]))**2  + b*setsb[2,:]**4  * np.sin(setsb[0,:])
    # # generate model output yCi = f(B_A)
    # iyc = np.empty((npara,nsets))*np.nan
    # for i in range(npara):
    #     tmpset = np.copy(setsb)
    #     tmpset[i,:] = setsa[i,:]
    #     iyc[i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + b*tmpset[2,:]**4  * np.sin(tmpset[0,:])
    # # generate model output yCi = f(A_B)
    # iyc2 = np.empty((npara,nsets))*np.nan
    # for i in range(npara):
    #     tmpset = np.copy(setsa)
    #     tmpset[i,:] = setsb[i,:]
    #     iyc2[i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + b*tmpset[2,:]**4  * np.sin(tmpset[0,:])

    # # Theoretical results: has to converge for n->Infinity to this values
    # vy = 0.5 + a**2/8.0 + b*np.pi**4/5.0+b**2*np.pi**8/18.0
    # theo_si  = np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0)/vy, (a**2/8.0)/vy, 0.0])
    # theo_sti = np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0+8.0/225.0*b**2*np.pi**8)/vy,
    #                      (a**2/8.0)/vy, (8.0/225.0*b**2*np.pi**8)/vy])
    # # Theoretical results
    # print('T :: si  =',astr(theo_si,3,pp=True))
    # #T :: si  = ['0.372' '0.000' '0.000']
    # print('T :: sti =',astr(theo_sti,3,pp=True))
    # #T :: sti = ['1.000' '0.000' '0.628']
    # #
    # # 3 positional arguments
    # isi1, isti1 = sobol_index(iya,iyb,iyc2)
    # print('S :: si  =',astr(isi1,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']
    # print('S :: sti =',astr(isti1,3,pp=True))
    # #S :: sti = ['1.014' '0.000' '0.654']

    # # 3 optional arguments
    # isi2, isti2 = sobol_index(ya=iya,yb=iyb,yc=iyc2)
    # print('S :: si  =',astr(isi2,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']

    # # 2 positional arguments
    # s2 = np.concatenate((iya, iyb, np.ravel(iyc2)))
    # ns = iya.size
    # isi3, isti3 = sobol_index(s2, ns)
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']

    # # 2 optional arguments
    # isi4, isti4 = sobol_index(s=s2, ns=ns)
    # print('S :: si  =',astr(isi4,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']

    # # 2 optional arguments and no STi output
    # isi5 = sobol_index(s=s2, ns=ns, si=True, sti=False)
    # print('S :: si  =',astr(isi5,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']

    # # 2 optional arguments and no Si output
    # isti5 = sobol_index(s=s2, ns=ns, si=False, sti=True)
    # print('S :: sti =',astr(isti5,3,pp=True))
    # #S :: sti = ['1.014' '0.000' '0.654']

    # # Method=Saltelli2008
    # s = np.concatenate((iya, iyb, np.ravel(iyc)))
    # isi3, isti3 = sobol_index(s, ns, method='Saltelli2008')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = ['0.367' '0.011' '0.016']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = [' 1.038' '-0.019' ' 0.641']

    # # Method=Homma1996
    # isi3, isti3 = sobol_index(s, ns, method='homma1996')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.356' ' 0.000' ' 0.005']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = [' 1.038' '-0.019' ' 0.641']

    # # Method=Mai2012
    # isi3, isti3 = sobol_index(s, ns, method='mai2012')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.353' ' 0.000' ' 0.005']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = ['1.037' '0.000' '0.647']

    # # Method=saltelli2010
    # isi3, isti3 = sobol_index(s2, ns, method='saltelli2010')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.348' ' 0.000' '-0.049']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = [' 0.984' ' 0.000' ' 0.633']

    # # Method=jansen1999
    # isi3, isti3 = sobol_index(s2, ns, method='jansen1999')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.319' ' 0.002' '-0.068']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = ['1.014' '0.000' '0.654']

    # # Method=mai2013
    # isi3, isti3 = sobol_index(s2, ns, method='mai2013')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = [' 0.984' ' 0.000' ' 0.633']

    # # Method=mai2014
    # isi3, isti3 = sobol_index(s2, ns, method='mai2014')
    # print('S :: si  =',astr(isi3,3,pp=True))
    # #S :: si  = [' 0.345' ' 0.000' '-0.049']
    # print('S :: sti =',astr(isti3,3,pp=True))
    # #S :: sti = ['1.014' '0.000' '0.654']

    # # ------------------------------------------------------------------------
    # #                                2D
    # #
    # # number of timepoints
    # ntime = 100
    # iya   = np.empty((ntime,nsets))*np.nan
    # iyb   = np.empty((ntime,nsets))*np.nan
    # iyc   = np.empty((ntime,npara,nsets))*np.nan
    # iyc2  = np.empty((ntime,npara,nsets))*np.nan
    # s     = np.empty((ntime,(npara+2)*nsets))*np.nan
    # s2    = np.empty((ntime,(npara+2)*nsets))*np.nan
    # ns    = nsets
    # theo_si  = []
    # theo_sti = []
    # for t in range(ntime):
    #     a = 1.0*t + 50.0
    #     b = 2.0
    #     iya[t,:] = np.sin(setsa[0,:])  + a*(np.sin(setsa[1,:]))**2  + b*setsa[2,:]**4  * np.sin(setsa[0,:])
    #     iyb[t,:] = np.sin(setsb[0,:])  + a*(np.sin(setsb[1,:]))**2  + b*setsb[2,:]**4  * np.sin(setsb[0,:])
    #     for i in range(npara):
    #         tmpset = np.copy(setsb)
    #         tmpset[i,:] = setsa[i,:]
    #         iyc[t,i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + (
    #                      b*tmpset[2,:]**4  * np.sin(tmpset[0,:]))
    #     for i in range(npara):
    #         tmpset = np.copy(setsa)
    #         tmpset[i,:] = setsb[i,:]
    #         iyc2[t,i,:] = np.sin(tmpset[0,:])  + a*(np.sin(tmpset[1,:]))**2  + (
    #                       b*tmpset[2,:]**4  * np.sin(tmpset[0,:]))
    #     s[t,:]  = np.concatenate((iya[t,:], iyb[t,:], np.ravel(iyc[t,:,:])))
    #     s2[t,:] = np.concatenate((iya[t,:], iyb[t,:], np.ravel(iyc2[t,:,:])))
    #     vy = 0.5 + a**2/8.0 + b*np.pi**4/5.0+b**2*np.pi**8/18.0
    #     theo_si  += [np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0)/vy, (a**2/8.0)/vy, 0.0])]
    #     theo_sti += [np.array([(0.5+b*np.pi**4/5.0+b**2*np.pi**8/50.0+8.0/225.0*b**2*np.pi**8)/vy,
    #                            (a**2/8.0)/vy, (8.0/225.0*b**2*np.pi**8)/vy])]

    # # 1D model output --> ya, yb are 2D and yc is 3D
    # isi1, isti1 = sobol_index(iya,iyb,iyc2)

    # # SI :: 1st time point
    # print('S :: si[t0]  =',astr(isi1[0,:],3,pp=True))
    # #S :: si[t0]  = [' 0.295' ' 0.147' '-0.056']
    # print('T :: si[t0]  =',astr(theo_si[0],3,pp=True))
    # #T :: si[t0]  = ['0.325' '0.127' '0.000']
    # # SI :: 3rd time point
    # print('S :: si[t2]  =',astr(isi1[2,:],3,pp=True))
    # #S :: si[t2]  = [' 0.293' ' 0.156' '-0.055']
    # print('T :: si[t2]  =',astr(theo_si[2],3,pp=True))
    # #T :: si[t2]  = ['0.321' '0.136' '0.000']
    # # STI :: 1st time point
    # print('S :: sti[t0] =',astr(isti1[0,:],3,pp=True))
    # #S :: sti[t0] = ['0.881' '0.126' '0.568']
    # print('T :: sti[t0] =',astr(theo_sti[0],3,pp=True))
    # #T :: sti[t0] = ['0.873' '0.127' '0.548']
    # # STI :: 3rd time point
    # print('S :: sti[t2] =',astr(isti1[2,:],3,pp=True))
    # #S :: sti[t2] = ['0.871' '0.134' '0.562']
    # print('T :: sti[t2] =',astr(theo_sti[2],3,pp=True))
    # #T :: sti[t2] = ['0.864' '0.136' '0.543']

    # # Mean
    # isi2, isti2, msi2, msti2 = sobol_index(ya=iya,yb=iyb,yc=iyc2, mean=True)
    # print('S :: si_m    =',astr(msi2,3,pp=True))
    # #S :: si_m    = [' 0.231' ' 0.377' '-0.037']
    # print('S :: sti_m   =',astr(msti2,3,pp=True))
    # #S :: sti_m   = ['0.643' '0.351' '0.415']
    # #
    # # Weighted mean
    # isi4, isti4, wsi2, wsti2 = sobol_index(s2, ns, wmean=True)
    # print('S :: si_w    =',astr(wsi2,3,pp=True))
    # #S :: si_w    = [' 0.224' ' 0.404' '-0.035']
    # print('S :: sti_w   =',astr(wsti2,3,pp=True))
    # #S :: sti_w   = ['0.616' '0.377' '0.397']

    # # 2 positional arguments
    # isi3, isti3 = sobol_index(s=s2, ns=ns)
    # # SI :: 3rd time point
    # print('S :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #S :: si[t2]  = [' 0.293' ' 0.156' '-0.055']
    # print('T :: si[t2]  =',astr(theo_si[2],3,pp=True))
    # #T :: si[t2]  = ['0.321' '0.136' '0.000']

    # # Method=Saltelli2008
    # isi3, isti3 = sobol_index(s, ns, method='saltelli2008')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = ['0.354' '0.130' '0.011']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = ['0.352' '0.139' '0.010']

    # # Method=Homma1996
    # isi3, isti3 = sobol_index(s, ns, method='homma1996')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.342' ' 0.118' '-0.002']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.340' ' 0.127' '-0.002']

    # # Method=Mai2012
    # isi3, isti3 = sobol_index(s, ns, method='Mai2012')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.336' ' 0.116' '-0.002']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.334' ' 0.125' '-0.002']

    # # Method=Saltelli2010
    # isi3, isti3 = sobol_index(s2, ns, method='saltelli2010')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.294' ' 0.142' '-0.031']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.291' ' 0.152' '-0.031']

    # # Method=Jansen1999
    # isi3, isti3 = sobol_index(s2, ns, method='jansen1999')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.284' ' 0.133' '-0.072']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.281' ' 0.143' '-0.072']

    # # Method=Mai2013
    # isi3, isti3 = sobol_index(s2, ns, method='Mai2013')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.290' ' 0.140' '-0.031']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.286' ' 0.149' '-0.030']

    # # Method=Mai2014
    # isi3, isti3 = sobol_index(s2, ns, method='mai2014')
    # print('Sal :: si[t0]  =',astr(isi3[0,:],3,pp=True))
    # #Sal :: si[t0]  = [' 0.295' ' 0.147' '-0.056']
    # print('Sal :: si[t2]  =',astr(isi3[2,:],3,pp=True))
    # #Sal :: si[t2]  = [' 0.293' ' 0.156' '-0.055']



#!/usr/bin/env python
import numpy as np
from pyjams.jams.dec2date import dec2date


def ustarflag(date, data, flags, isday, outdir, min_thresh=0.1, nboot=100,
              ustar_v=2, plot=False):
    '''
    Friction velocity flagging for Eddy Covariance data. Assesses the threshold
    of friction velocity (ustar) below which a reduction in CO2 flux correlates
    with ustar and returns a flag here. Originally coded by Tino Rau.


    Definition
    ----------
    ustarflag(date, isday, data, flags, outdir, min_thresh=0.1, nboot=100,
              ustar_v=2, plot=False):


    Input
    -----
    date        np.array(N), julian date
    data        np.array(N,3), data array with CO2 flux (Fco2 [mumol/m2s),
                friction velocity (ustar [m/s]) and air temperature (T [degC])
    inflag      np.array(N,3), dtype=int, quality flag of data, 0 where data is
                good
    isday       np.array(N), dtype=bool, True where it is day and False where
                it is night
    outdir      str, path of the output folder


    Optional Input
    --------------
    min_thresh  float, minimum ustar threshold, recommendation: 0.1 for
                overstorey towers, 0.01 for understorey towers (default: 0.1)
    nboot       int, number of boot straps (default: 100)
    ustar_v     int, value which shall be returned when a value is below the
                ustar threshold (default: 2)
    udef        int/float, missing value of data (default: -9999) NaN values are
                excluded from computations anyhow.
    plot        bool, if True data and spikes are plotted (default: False)


    Output
    ------
    flag_out    np.array(N), flag array where everything is 0 except where
                values fall below ustar threshold, there it is ustar_v


    Restrictions
    ------------
    - works ONLY for a data set of ONE FULL year
    - works ONLY for half hourly time steps


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''

    ############################################################################
    # fixed parameters
    nperiods   = 4    # numer of seasons per year
    ntclass    = 6    # number of temperature classes
    corrthresh = 0.3  # correlation coefficient
    nuclass    = 20   # number of u* classes
    upercent   = 95.  # percentage for average comparison
    udef       =-9999 # undef of output

    ############################################################################
    # data and flags
    Fco2       = data[:,0]
    ustar      = data[:,1]
    T          = data[:,2]
    flag_Fco2  = flags[:,0]
    flag_ustar = flags[:,1]
    flag_T     = flags[:,2]

    years, months, days, hours, mins, sc = dec2date(date, fulldate=True)
    nmons = 12/nperiods
    yrmin = np.min(years)
    yrmax = np.max(years)
    nyears = yrmax - yrmin ######## + 1 works only for one full year of data
    if (nyears>1) | (np.unique(months).size<12):
        raise ValueError('ustarflag: only one full year of data can be processed')
    flag = (flag_Fco2==0)&(~isday)&(flag_T==0)&(flag_ustar==0)

    ############################################################################
    # prepare bootstrapping
    threshs = np.zeros((nboot,nyears))
    seasonal_threshs = np.zeros((nperiods*nboot,nyears))

    ############################################################################
    # calculate thresholds
    for y in xrange(yrmin, yrmax): ######## + 1 works only for one full year of data
        #print 'Year ', y
        periods= np.unique(months[np.where(years==y)])[0:-1:nmons]
        nperiods = len(np.unique(months[np.where(years==y)]))/nmons
        out_threshs = np.array([])
        nstepsyear = len(np.where(years==y)[0])
        #print nstepsyear

        ########################################################################
        # start bootstrapping
        for r in xrange(nboot):
            range_min = np.min(np.where(years==y)[0])
            range_max = np.max(np.where(years==y)[0])
            rand      = np.random.randint(0,nstepsyear,size=nstepsyear)
            T_b       = T[range_min+rand]
            flag_b    = flag[range_min+rand]
            years_b   = years[range_min+rand]
            months_b  = months[range_min+rand]
            ustar_b   = ustar[range_min+rand]
            Fco2_b    = Fco2[range_min+rand]

            # flag the created bootstrap-arrays
            T_f      = np.extract(flag_b, T_b)
            months_f = np.extract(flag_b, months_b)
            years_f  = np.extract(flag_b, years_b)
            ustar_f  = np.extract(flag_b, ustar_b)
            Fco2_f   = np.extract(flag_b, Fco2_b)

            # container for thresholds of the seasons of a year
            period_threshs = np.array([])

            ####################################################################
            # loop over seasons
            for m in periods:
                period_flag = (months_f>=m) & (months_f<m+nmons)

                # flag data --> just night_time and unflagged data is left
                T_m     = np.ma.array(T_f,     mask=~period_flag, fill_value=udef)
                ustar_m = np.ma.array(ustar_f, mask=~period_flag, fill_value=udef)
                Fco2_m  = np.ma.array(Fco2_f,  mask=~period_flag, fill_value=udef)

                # sort the array according to the temperature
                ii      = np.ma.argsort(T_m)
                T_m     = T_m[ii]
                ustar_m = ustar_m[ii]
                Fco2_m  = Fco2_m[ii]

                # separate into 6 temperature classes with
                # equal size according to quantiles
                class_width = np.ma.count(T_m)/ntclass
                ntlen       = T_m.size

                # container for the thresholds of this period
                class_threshs = np.array([])

                ################################################################
                # loop for every temperature class
                for i in xrange(ntclass):
                    T_temp    = T_m[i*class_width:
                                    np.min([(i+1)*class_width, ntlen-1])]
                    u_temp    = ustar_m[i*class_width:
                                        np.min([(i+1)*class_width, ntlen-1])]
                    Fco2_temp = Fco2_m[i*class_width:
                                       np.min([(i+1)*class_width, ntlen-1])]

                    # calculate r
                    #try: #AP
                    r_abs = np.abs(np.corrcoef(T_temp,u_temp)[0,1])
                    #except IndexError: #AP
                    #    print 'IndexError'#AP
                    #    r_abs = 1 #AP

                    # neglect threshold, if correlation is strong
                    # online gap tool uses 0.3 as r-threshold,
                    # original papale paper: 0.4
                    if r_abs >= corrthresh:
                        continue

                    # sort the data of the temperature class according
                    # to ustar
                    ii        = np.ma.argsort(u_temp)
                    T_temp    = T_temp[ii]
                    u_temp    = u_temp[ii]
                    Fco2_temp = Fco2_temp[ii]

                    # set the class width of the ustar classes; class
                    # with equal size??
                    u_class_width = np.ma.count(T_temp)/nuclass
                    nulen         = T_temp.size

                    ############################################################
                    # loop over all ustar classses-1
                    # threshold = if class average > 99% of average of
                    # all classes above
                    # loop over all ustar classes
                    for u in xrange(nuclass-1):
                        T_u    = T_temp[u*u_class_width:
                                        np.min([(u+1)*u_class_width,nulen-1])]
                        u_u    = u_temp[u*u_class_width:
                                        np.min([(u+1)*u_class_width,nulen-1])]
                        Fco2_u = Fco2_temp[u*u_class_width:
                                           np.min([(u+1)*u_class_width,nulen-1])]

                        rest_Fco2 = Fco2_temp[np.min([(u+1)*u_class_width,nulen-1]):]
                        rest_avg  = np.ma.mean(rest_Fco2)
                        avg       = np.ma.mean(Fco2_u)

                        if np.abs(avg) >= np.abs(upercent/100.*rest_avg):
                            class_threshs = np.append(class_threshs, np.ma.mean(u_u))
                            break

                # median of thresholds of all temperature classes is
                # set as the threshold of the period
                if ~np.isnan(np.ma.median(class_threshs)):
                    #print r
                    period_threshs = np.append(period_threshs,
                                               np.ma.median(class_threshs))
                else:
                    #print 'All are nans',class_threshs
                    period_threshs = np.append(period_threshs, udef)

            # threshholds of all periods
            out_threshs = np.append(out_threshs,period_threshs)
            # if there are values in period_thresholds take the maximum,
            # else take 90th percentil
            if period_threshs == []:
                threshs[r,y-yrmin] = np.sort(ustar_f)[np.len(ustar_f)*9./10.]
            else:
                threshs[r,y-yrmin] = np.max(period_threshs)

        seasonal_threshs[0:nperiods*nboot,y-yrmin] = out_threshs

    ############################################################################
    # set the flags

    # assign calculated threshold:
    # for each year separately,
    # for night and day-time data
    # take the median of the nboot bootstrapped thresholds for each year
    # "out_threshs" used for export and later plotting

    flag_out = np.zeros_like(flag_Fco2, dtype=np.int)

    for y in range(yrmin, yrmax): ######## + 1 works only for one full year data
        if np.any(threshs[:,y-yrmin] > 10):
            print('Hu', y, threshs[:,y-yrmin])
        med = np.median(threshs[:,y-yrmin])
        if (med < min_thresh):
            med = min_thresh
        elif np.isnan(med):
            print('NaN-WARNING!')
            ii = np.argsort(ustar[years==y])
            oo =np.where(flag_Fco2[ii]==0)[0]
            flag_out[years==y][ii[oo][::np.int(len(oo)*0.9)]]\
                    += ustar_v
        else:
            ii = (years==y).flatten()
            yy = np.where((ustar[ii]<med) & (flag_Fco2[ii]==0))
            uu = np.clip(np.unique(np.concatenate((yy[0], yy[0]+1))), 0,
                         len(flag_Fco2[(years==y).flatten()])-1)
            flag_out[np.where(years==y)[0][uu]]+=ustar_v

            #print 'Year: %i, u_star thresh: %5.3f'%(y, med)

            ####################################################################
            # plot
            if plot:
                import matplotlib.pyplot as plt
                import matplotlib.backends.backend_pdf as pdf
                fig1 = plt.figure(1)
                sub1 = fig1.add_subplot(111)
                sub1.plot(ustar[np.where(flag & (years==y).flatten())],\
                          Fco2[np.where(flag & (years==y).flatten())], 'bo')
                sub1.axvline(x=med, linewidth=0.75, color='r')
                plt.ylabel('F CO2')
                plt.xlabel('u_star')
                plt.title('u_star thresh: %5.3f'%med)
                plt.show()

                pp1 = pdf.PdfPages(outdir+'/ustar.pdf')
                fig1.savefig(pp1, format='pdf')
                pp1.close()

    ############################################################################
    # save output
    np.savetxt('%s/u_star_thresh.csv' %outdir, threshs,
               fmt='%4.4f', delimiter=',')
    np.savetxt('%s/u_star_thresh_seasonal.csv' %outdir, seasonal_threshs,
               fmt='%4.4f', delimiter=',')

    return flag_out

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
import numpy as np


__all__ = ['grid_mid2edge']


def grid_mid2edge(lon, lat):
    """
        2D arrays of longitude and latitude grid edges from grid midpoints.


        Definition
        ----------
        def grid_mid2edge(lon, lat):


        Input
        -----
        lon        1D (nlon) or 2D array (nlat,nlon) of longitude grid midpoints
        lat        1D (nlat) or 2D array (nlat,nlon) of latitude grid midpoints


        Optional Input
        --------------
        None


        Output
        ------
        lon2, lat2 = 2D arrays (nlat+1,nlon+1) of longitude and latitude grid edges.


        Restrictions
        ------------
        Longitudes and latitudes are interpolated as on a plane, not a sphere.
        However, rotated grids should be possible.


        Examples
        --------
        >>> from jams.autostring import astr
        >>> nlon = 36
        >>> nlat = 18

        Lon -180 - +180, Lat -90 - +90, i.e. S-N
        >>> lon = np.arange(nlon)*(360./nlon) - 175. # -180 - +180
        >>> lat = np.arange(nlat)*(180./nlat) - 85.  # S-N
        >>> lonh, lath = grid_mid2edge(lon, lat)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['-180' ' 180']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        ['-90' ' 90']

        Lon -180 - +180, Lat +90 - -90, i.e. N-S
        >>> lat = lat[::-1]                          # N-S
        >>> lonh, lath = grid_mid2edge(lon, lat)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['-180' ' 180']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        [' 90' '-90']

        Lon 0 - 360, Lat +90 - -90, i.e. N-S
        >>> lon += 180.                              # 0-360
        >>> lonh, lath = grid_mid2edge(lon, lat)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['  0' '360']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        [' 90' '-90']

        Rolled axis for different map projections, Lat +90 - -90, i.e. N-S
        >>> lon  = np.roll(lon, nlon//3)             # 245-235
        >>> lonh, lath = grid_mid2edge(lon, lat)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['240' '240']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        [' 90' '-90']
        >>> lon -= 180.                              # 65-55
        >>> lonh, lath = grid_mid2edge(lon, lat)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['60' '60']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        [' 90' '-90']

        2D input
        >>> lon = np.arange(nlon)*(360./nlon) - 175. # -180 - +180
        >>> lat = np.arange(nlat)*(180./nlat) - 85.  # S-N
        >>> lon2, lat2 = np.meshgrid(lon,lat)
        >>> lonh, lath = grid_mid2edge(lon2, lat2)
        >>> print(astr([lonh[0,0],lonh[-1,-1]]))
        ['-180' ' 180']
        >>> print(astr([lath[0,0],lath[-1,-1]]))
        ['-90' ' 90']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Oct 2014
    """
    #
    # Check input
    do2 = np.ndim(lon)
    da2 = np.ndim(lat)
    assert (do2 >=1) & (do2 <= 2), 'longitudes have to be either 1D or 2D arrays.'
    assert (da2 >=1) & (da2 <= 2), 'latitudes have to be either 1D or 2D arrays.'
    assert do2 == da2, 'longitudes and latitudes have to be either 1D or 2D arrays; no mixture possible.'
    #
    # Make 2D
    if do2 == 1:
        lon2, lat2 = np.meshgrid(lon,lat) # have sizes 
    else:
        lon2 = lon # no copy
        lat2 = lat
    #
    # N-S or S-N
    isns = True                                 # descending latitudes
    if np.min(np.diff(lat2)) > 0.: isns = False # ascending latitudes
    #
    # 0-360 0r -180-+180
    is360 = True          #    0 -  360
    if np.any(lon2 < 0.): # -180 - +180
        lon2  = np.copy(lon2) + 180. # do all calculations in 0-360
        is360 = False
    #
    # out arrays
    nlat = lon2.shape[0]
    nlon = lon2.shape[1]
    lonh = np.empty((nlat+1, nlon+1))
    lath = np.empty((nlat+1, nlon+1))
    # Edge points in interior = mean of 4 surrounding grid boxes
    lonh[1:-1,1:-1] = 0.25*(lon2[0:-1,0:-1] + # upper left
                            lon2[0:-1,1:] +   # upper right
                            lon2[1:,0:-1] +   # lower left
                            lon2[1:,1:])      # lower right
    lath[1:-1,1:-1] = 0.25*(lat2[0:-1,0:-1] +
                            lat2[0:-1,1:] +   # same as lon
                            lat2[1:,0:-1] +
                            lat2[1:,1:])
    # left column w/o corners = left - half distance
    lonh[1:-1,0]  = 0.75*lon2[0:-1,0] + 0.75*lon2[1:,0] - 0.25*lon2[0:-1,1] - 0.25*lon2[1:,1]
    lath[1:-1,0]  = 0.5*(lat2[0:-1,0] + lat2[1:,0])
    # right column w/o corners = right + half distance
    lonh[1:-1,-1] = 0.75*lon2[0:-1,-1] + 0.75*lon2[1:,-1] - 0.25*lon2[0:-1,-2] - 0.25*lon2[1:,-2]
    lath[1:-1,-1] = 0.5*(lat2[0:-1,-1] + lat2[1:,-1])
    # upper row w/o corners = up + half distance
    lonh[0,1:-1]  = 0.5*(lon2[0,0:-1] + lon2[0,1:])
    lath[0,1:-1]  = 0.75*lat2[0,0:-1] + 0.75*lat2[0,1:] - 0.25*lat2[1,0:-1] - 0.25*lat2[1,1:]
    # lower row w/o corners = low - half distance
    lonh[-1,1:-1]  = 0.5*(lon2[-1,0:-1] + lon2[-1,1:])
    lath[-1,1:-1]  = 0.75*lat2[-1,0:-1] + 0.75*lat2[-1,1:] - 0.25*lat2[-2,0:-1] - 0.25*lat2[-2,1:]
    # corners = midpoint plus or minus dist to opposite corner
    lonh[0,0]   = lon2[0,0]   - (lonh[1,1] - lon2[0,0])     # upper left
    lath[0,0]   = lat2[0,0]   - (lath[1,1] - lat2[0,0])
    lonh[0,-1]  = lon2[0,-1]  - (lonh[1,-2] - lon2[0,-1])   # upper right
    lath[0,-1]  = lat2[0,-1]  - (lath[1,-2] - lat2[0,-1])
    lonh[-1,0]  = lon2[-1,0]  - (lonh[-2,1] - lon2[-1,0])   # lower left
    lath[-1,0]  = lat2[-1,0]  - (lath[-2,1] - lat2[-1,0])
    lonh[-1,-1] = lon2[-1,-1] - (lonh[-2,-2] - lon2[-1,-1]) # lower right
    lath[-1,-1] = lat2[-1,-1] - (lath[-2,-2] - lat2[-1,-1])

    # lon can be > 360
    # this makes 360=0 and 180=-180: stay with 360 and 180
    #    lonh %= 360
    if np.any(lonh > 360.):
        lonh = np.where(lonh > 360., lonh % 360., lonh)
    
    # return to -180-+180
    if not is360:
        lonh -= 180.

    return lonh, lath


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from jams.autostring import astr
    # nlon = 36
    # nlat = 18
    # lon = np.arange(nlon)*(360./nlon) - 175. # -180 - +180
    # lat = np.arange(nlat)*(180./nlat) - 85.  # S-N
    # lonh, lath = grid_mid2edge(lon, lat)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['-180' ' 180']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    ['-90' ' 90']
    # lat = lat[::-1]                          # N-S
    # lonh, lath = grid_mid2edge(lon, lat)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['-180' ' 180']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    [' 90' '-90']
    # lon += 180.                              # 0-360
    # lonh, lath = grid_mid2edge(lon, lat)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['  0' '360']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    [' 90' '-90']
    # lon2, lat2 = np.meshgrid(lon,lat)
    # lonh, lath = grid_mid2edge(lon2, lat2)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['  0' '360']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    [' 90' '-90']
    # lon  = np.roll(lon, nlon//3)             # 245-235
    # lonh, lath = grid_mid2edge(lon, lat)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['240' '240']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    [' 90' '-90']
    # lon -= 180.                              # 65 - 55
    # lonh, lath = grid_mid2edge(lon, lat)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # #    ['60' '60']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # #    [' 90' '-90']
    # # 2D
    # lon = np.arange(nlon)*(360./nlon) - 175. # -180 - +180
    # lat = np.arange(nlat)*(180./nlat) - 85.  # S-N
    # lon2, lat2 = np.meshgrid(lon,lat)
    # lonh, lath = grid_mid2edge(lon2, lat2)
    # print(astr([lonh[0,0],lonh[-1,-1]]))
    # # ['-180' ' 180']
    # print(astr([lath[0,0],lath[-1,-1]]))
    # # ['-90' ' 90']



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def rolling(x, win):
    """
        Uses binary strides to reshape an array in a "rolling window" style
        to perform any function on a rolling window without for loop


        Definition
        ----------
        def rolling(x, win):


        Input
        -----
        x         1D np.array(N), array to be rehaped
        win       int, window size


        Output
        ------
        x2D,      2D np.array(N-(win+1),win), x reshaped in windows


        Restrictions
        ------------
        The method is only applicable for 1D numpy arrays.


        References
        ----------
        This routine is taken from Alex Rogozhnikov 
        (http://arogozhnikov.github.io/)


        Example
        --------
        >>> x = np.arange(10)

        # get windowed array with window size 3
        >>> win = 3
        >>> print(rolling(x, win).shape)
        (8, 3)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Dec 2016
    """
    # test input
    assert x.ndim == 1,          'x must be 1D'
    assert isinstance(win, int), 'win must be integer'

    s   = x.strides[0]
    x2D = np.lib.stride_tricks.as_strided(x, shape=[x.size - win+ 1, win], strides=[s, s])
    return x2D

    
if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
import scipy.stats as stats


__all__ = ['lhs']


def lhs(dist, param, nsample):
    """
    Latin Hypercube Sampling of any distribution without correlations
    after Stein (1987).


    Definition
    ----------
    def lhs(dist, param, nsample):


    Input
    -----
    dist
        random number generator (list) from scipy.stats such as stats.norm,
        stats.beta, etc.
    param
        tuple of parameters as required for dist
    nsample
        number of samples per parameter


    Output
    ------
    Latin hypercube Sample array of [size(nsample),nsample]


    Restrictions
    ------------
    No correlations between parameters possible.


    References
    ----------
    Stein, M. 1987. Large Sample Properties of Simulations Using
        Latin Hypercube Sampling. Technometrics 29:143-151


    Examples
    --------
    >>> import numpy as np
    >>> import scipy.stats as stats
    >>> # seed for reproducible results in doctest
    >>> np.random.seed(1)
    >>> dist = [stats.norm, stats.uniform] # for uniform (min, max-min)
    >>> pars = [(50,2),(1,5)]
    >>> c    = lhs(dist, pars, 20)
    >>> from autostring import astr
    >>> print(astr(c[0:2,0:4],3,pp=True))
    [['52.822' '51.956' '46.710' '50.585']
     [' 4.950' ' 2.492' ' 2.078' ' 4.673']]

    >>> np.random.seed(1)
    >>> dist = [stats.norm]
    >>> pars = [(50,2)]
    >>> c    = lhs(dist, pars, 20)
    >>> print(c.shape)
    (1, 20)
    >>> print(astr(c[0,0:4],3,pp=True))
    ['51.171' '48.562' '51.683' '50.585']

    >>> np.random.seed(1)
    >>> dist = stats.norm
    >>> pars = (50,2)
    >>> c    = lhs(dist, pars, 20)
    >>> print(c.shape)
    (20,)
    >>> print(astr(c[0:4],3,pp=True))
    ['51.171' '48.562' '51.683' '50.585']


    History
    -------
    Written,  Matthias Cuntz, May 2012
                  - combination of Matlab routines of Budiman (2003)
                    and Python routines of Flavio Codeco Coelho (2008)
    Modified, Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Nov 2016 - preserve shape <- nodim
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # Check input
    if not isinstance(dist, (list, tuple)):
        nodim = True
        dist  = [dist]
        param = [param]
    else:
        nodim = False
        assert len(dist) == len(param)
    ndist = len(dist)

    # LHS
    ran    = np.random.uniform(0., 1., (ndist, nsample))
    lhsout = np.empty((ndist, nsample))
    for j, d in enumerate(dist):
        if not isinstance(d, (stats.rv_discrete, stats.rv_continuous)):
            raise TypeError('dist is not a scipy.stats distribution object.')
        # force type to float for sage compatibility
        pars = tuple([float(k) for k in param[j]])
        idx = np.array(np.random.permutation(nsample), dtype=float)
        p   = (idx+ran[j, :])/float(nsample)  # probability of cdf
        lhsout[j, :] = d(*pars).ppf(p)        # inverse of cdf

    if nodim:
        return lhsout[0, :]
    else:
        return lhsout


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import matplotlib.pyplot as plt
    # dist = [stats.norm, stats.uniform]
    # pars = [(50,2),(1,5)]
    # c    = lhs(dist, pars, 20000)

    # plt.figure()
    # plt.hist(c[0,:])

    # plt.figure()
    # plt.hist(c[1,:])

    # dist = [stats.uniform, stats.uniform]
    # pars = [(50,2),(1,5)]
    # c    = lhs(dist, pars, 20000)

    # plt.figure()
    # plt.plot(c[0,:],c[1,:],'ko',markersize=1.0)
    # plt.show()



#!/usr/bin/env python  # -*- coding: utf-8 -*-
import numpy as np
from netcdf4 import NcDataset


def concatDimension(fnames, dim, dimvar=None, outfname=None, **kwargs):
    """
    Arguments:
    ----------
    datasets (List[str]):          datasets to concatenate
    dim (String):                  name of the dimension to concatenate
    dimvar (Optional[String]):     name of the variable defining values
                                   for the concatenation dimension,
                                   default: dimvar = dim
    outfname (Optional[str]):      file name of an output dataset
    kwargs (Optional[Any]):        paramaters to pass to the createVariable
                                   Method of NcDataset


    Return
    ------
    Optional[netcdf4.NcDataset]

    Purpose
    -------
    Concatenate the given datasets into a new one along the given dimension
    'dim'. Values of the dimensions need to be given in 'dimvar' (default
    variable name is the same as the concatenation dimension).
    If 'outfname' is given the concatenation is file based, otherwise an in-memory
    dataset will be returned.
    """

    def getDimVarValues(fnames, vardim):
        vals = []
        types = []
        for fname in fnames:
            with NcDataset(fname, "r") as nc:
                # TODO:
                # check if values are given, otherwise make up
                # some fake data simpling starting at 0
                data = nc.variables[vardim][:]
                types.append(data.dtype)
                vals.extend(data.tolist())
        return np.unique(vals).astype(dtype=np.find_common_type(types, []))

    def getCommonVarTypes(fnames):
        vars = {}
        for fname in fnames:
            with NcDataset(fname, "r") as nc:
                for vname, var in nc.variables.items():
                    try:
                        vars[vname].append(var.dtype)
                    except KeyError:
                        vars[vname] = [var.dtype]
        return {
            vname: np.find_common_type(vtypes, [])
            for vname, vtypes in vars.items()}

    if not dimvar:
        dimvar = dim

    dimvals = getDimVarValues(fnames, dimvar)
    dtypes = getCommonVarTypes(fnames)

    # the new concat dimension
    out = NcDataset(outfname, "w")
    out.createDimension(dim, None, fail=False)
    var = out.createVariable(dimvar, dimvals.dtype, (dim,), fail=False, **kwargs)
    var[:] = dimvals

    for i, fname in enumerate(fnames):

        with NcDataset(fname, "r") as nc:

            out.copyDimensions(nc.dimensions, skip=out.dimensions, fix=True)
            out.copyAttributes(nc.attributes)

            for vname, var in nc.variables.items():

                outvar = out.copyVariable(
                    var, dtype=dtypes[vname], data=False, fail=False, **kwargs)

                # build up the indices
                if dim in var.dimensions:
                    # idx = [dimvals.index(v) for v in nc.variables[dimvar]]
                    idx = [np.where(dimvals == v)[0][0] for v in nc.variables[dimvar]]
                    slices = [slice(None,)] * len(var.dimensions)
                    slices[var.dimensions.index(dim)] = idx
                else:
                    # writing them again is stupid...
                    slices = slice(None)

                data = var[:]

                # this seems to be necessary, because netCDF4 sometimes comes
                # up with a masked without a explicit user definition
                # (i.e. no fill_value, valid_range, whatever is set)
                if isinstance(data, np.ma.MaskedArray):
                    data = data.data
                outvar[slices] = data

    if outfname is None:
        return out
    out.close()



#!/usr/bin/env python
import numpy as np
import os as os
from pyjams.jams.date2dec import date2dec
from pyjams.jams.dec2date import dec2date
import re


def timestepcheck(indir, pat, outfile, begin, end, numhead=1, timeint=30,
                  format='ascii', empty='-9999', delimiter=',', skiprows=0):
    '''
    Checks ascii data files with first column being a time stamp in ascii
    of eng style for correctness. If for a given time interval time steps
    are missing, they will will be filled with the correct time stamp and
    empty values. A beginning and end for the output files can be set to which
    data should be kept or filled. Pat defines the file name pattern to consider
    in the check. Multiple files, which match the pat are concatenated.


    Definition
    ----------
    timestepcheck(indir, pat, outfile, begin, end, numhead=1, timeint=30,
                  format='ascii', empty='-9999', delimiter=',', skiprows=0):


    Input
    -----
    indir       str, path of the folder where the input files are
    pat         str, name or regular expression of the input files
    outfile     str, path and name of the output file
    begin       str, start time of the output file, must be in the same format
                as time stamps in the input files
    end         str, end time of the output file, must be in the same format
                as time stamps in the input files


    Optional Input
    --------------
    numhead     int, number of header lines in the input files (default: 1)
    timeint     int, time interval of the input file in minutes (default: 30)
    format      str, format of time stamps in input files. 'ascii' or 'eng' is
                possible (default: 'ascii')
    empty       str, value for missing values (default: '-9999')
    delimiter   str, delimiter of the input files (default: ',')
    skiprows    int, rows to skip in input files, e.g. logger fuzzle before
                actual data header starts (default: 0)


    Output
    ------
    outfile     file with missing time steps filled with empty values cut from
                begin to end


    Restrictions
    ------------
    TODO: tested thoroughly only for timeint=30
    TODO: more bad value checks can be included, see sternchen and tuedelchen


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''
    ###########################################################################
    # time interval list
    interval = range(0,60,timeint)
    #jdint    = date2dec(yr=-4712,mo=1,dy=1,hr=12,mi=timeint)
    #jdmin    = date2dec(yr=-4712,mo=1,dy=1,hr=12,mi=0,sc=30)# (=precision)
    jdint    = date2dec(yr=1,mo=1,dy=1,hr=12,mi=timeint) % 1
    jdmin    = date2dec(yr=1,mo=1,dy=1,hr=12,mi=0,sc=30) % 1# (=precision)
    if format == 'ascii':
        jdbegin  = date2dec(ascii = np.array([begin]))
        jdend    = date2dec(ascii = np.array([end]))
    elif format == 'eng':
        jdbegin  = date2dec(eng = np.array([begin]))
        jdend    = date2dec(eng = np.array([end]))

    ###########################################################################
    # reading input directory
    pat = re.compile(pat)
    new = True

    filelist = os.listdir(indir)
    for file in filelist:
        if re.search(pat, file):
            if new:
                data = np.loadtxt('./%s/%s'%(indir, file), dtype='|S21',\
                                  delimiter=delimiter, skiprows=skiprows)
                if data.shape[0] == 0:
                    print('Warning: File %s is empty!' %(file))
                else:
                    if np.shape(data.shape)[0] == 1:
                        data = data.reshape((1,-1))
                    new = False
            else:
                add_data = np.loadtxt('./%s/%s'%(indir, file), dtype='|S21',
                                      delimiter=delimiter,
                                      skiprows=numhead+skiprows)
                if add_data.shape[0] == 0:
                    print('Warning: File %s is empty!' %(file))
                elif np.shape(add_data.shape)[0] == 1:
                    add_data = add_data.reshape((1,-1))
                    data = np.append(data, add_data, 0)
                else:
                    data = np.append(data, add_data, 0)

    ###########################################################################
    # sternchen check :-D
    # replace with regular expression check
    data[data=='***********'] = empty #!!! uberprufen auf lange und re
    data[data=='********'] = empty #!!! uberprufen auf lange und re
    data[data=='*********'] = empty

    ###########################################################################
    # tuedelchen check :-D
    # replace with regular expression check
    if data[numhead,0][0] == '"':
        data[numhead:,0] = np.array([x[1:-1] for x in data[numhead:,0]])

    ###########################################################################
    # "NAN" check :-D
    # replace with regular expression check
    data[data=='"NAN"'] = empty #!!! uberprufen auf lange und re

    ###########################################################################
    # leerzeilen check
    blankline = np.where(data[0:2,0]=='')[0]
    data = np.delete(data, blankline, 0)

    ###########################################################################
    # missing values check
    data[data==''] = empty
    data[data=='""'] = empty

    columns = np.shape(data)[1]-1

    ###########################################################################
    # calculate julian date
    if format == 'ascii':
        import time
        jd = date2dec(ascii = data[numhead:,0])
    elif format == 'eng':
        jd = date2dec(eng = data[numhead:,0])

    ###########################################################################
    # wrong time stamp check
    diff = jd[1:] - jd[:-1]
    minute = np.array([x.split()[1][3:5] for x in data[numhead:,0]]).astype(int)
    nii    = np.nonzero(~np.in1d(minute, interval))[0]
    ts     = np.nonzero(np.less(diff, jdint-jdmin))[0]
    wrong  = np.unique(np.append(nii, ts))
    if data.shape[0]-numhead-2 in wrong:
        wrong = np.append(wrong, [data.shape[0]-numhead-1], 0)
    delete = []

    for i in wrong:
        print('\nHERE IS SOMETHING WRONG:\n')
        print('BOF' if numhead+i-2<0 else data[numhead+i-2,:4])
        print('BOF' if numhead+i-1<0 else data[numhead+i-1,:4])
        print('-----------------------------------------------')
        print(data[numhead+i,:4])
        print('-----------------------------------------------')
        print('EOF' if numhead+i+1>=np.shape(data)[0] else data[numhead+i+1,:4])
        print('EOF' if numhead+i+2>=np.shape(data)[0] else data[numhead+i+2,:4])

        do = raw_input("\n(d)elete entry, (s)et to empty, (t)ype in date, (i)gnore: ")

        if   do == 'd':
            delete += [numhead+i]
        elif do == 's':
            data[numhead+i,1:] = empty
        elif do == 't':
            newdate = str(raw_input("\nreplace with: "))
            data[numhead+i,0] = newdate
#            newmin = str(raw_input("\n%s"%(data[numhead+i,0][:-2])))
#            data[numhead+i,0] = data[numhead+i,0][:-2] + newmin
        elif do == 'i':
            pass

    data = np.delete(data, delete, 0)

    ###########################################################################
    # calculate julian date again
    if format == 'ascii':
        jd = date2dec(ascii = data[numhead:,0])
    elif format == 'eng':
        jd = date2dec(eng = data[numhead:,0])

    ###########################################################################
    # check time step
    diff = jd[1:] - jd[:-1]
    ingap = np.where(np.greater(diff, jdint+jdmin))[0]
    nugap = np.rint((diff[ingap]/jdint)-1)

    ###########################################################################
    # insert missing time steps
    for i in range(np.size(ingap))[::-1]:
        where = np.ones(int(nugap[i]), dtype=int)*(ingap[i]+1+numhead)
        if format == 'ascii':
            span  = np.arange(1,nugap[i]+1)*jdint + jd[ingap[i]]
            what  = dec2date(span.astype('|S16').astype(float), ascii=True)
        elif format == 'eng':
            span  = np.arange(1,nugap[i]+1)*jdint + jd[ingap[i]]
            what  = dec2date(span.astype('|S16').astype(float), eng=True)
        what    = np.array([x[:-3] for x in what])

        miss    = np.empty((int(nugap[i]),columns), dtype='|S11')
        miss[:] = empty
        what    = np.append(np.reshape(what, (-1,1)), miss, 1)
        data    = np.insert(data, where, what, 0)

    ###########################################################################
    # fill/cut up/off beginning and end
    start = np.where(data[:,0]==begin)[0]
    if start == numhead:
        pass
    elif start > numhead:
        data = np.delete(data, np.arange(numhead, start), 0)
    else:
        if format == 'ascii':
            tofill = int((date2dec(ascii = data[numhead,0]) - jdbegin)/jdint)
            span   = np.arange(0,tofill)*jdint + jdbegin # tofill+1 weggenommen!!!
            what   = dec2date(span.astype('|S16').astype(float), ascii=True)
        elif format == 'eng':
            tofill = int((date2dec(eng = data[numhead,0]) - jdbegin)/jdint)-1
            span   = np.arange(0,tofill)*jdint + jdbegin # tofill+1 weggenommen!!!
            what   = dec2date(span.astype('|S16').astype(float), eng=True)

        what  = np.array([x[:-3] for x in what])
        where = np.ones(tofill, dtype=int)*numhead # tofill+1 weggenommen!!!

        #if doidate:
        #    miss    = np.empty((tofill,columns-2), dtype='|S11') # tofill+1 weggenommen!!!
        #else:
        miss    = np.empty((tofill,columns), dtype='|S11') # tofill+1 weggenommen!!!
        miss[:] = empty
        what    = np.append(np.reshape(what, (-1,1)), miss, 1)
        data    = np.insert(data, where, what, 0)

    stop = np.where(data[:,0]==end)[0]
    maxind = np.shape(data)[0]-1
    if stop == maxind:
        pass
    elif stop < maxind:
        #data = data[:stop+1,:]
        data = data[:stop[0]+1,:]
    else:
        if format == 'ascii':
            tofill = int((jdend - date2dec(ascii = data[-1,0]))/jdint)
            span   = np.arange(1,tofill+1)*jdint + date2dec(ascii = data[-1,0])
            what   = dec2date(span.astype('|S16').astype(float), ascii=True)
        elif format == 'eng':
            tofill = int((jdend - date2dec(eng = data[-1,0]))/jdint)
            span   = np.arange(1,tofill+1)*jdint + date2dec(eng = data[-1,0])
            what   = dec2date(span.astype('|S16').astype(float), eng=True)

        what  = np.array([x[:-3] for x in what])
        #if doidate:
        #    miss    = np.empty((tofill,columns-2), dtype='|S11')
        #else:
        miss    = np.empty((tofill,columns), dtype='|S11')
        miss[:] = empty
        what    = np.append(np.reshape(what, (-1,1)), miss, 1)
        data    = np.append(data, what, 0)

    ###########################################################################
    # save data to txt file
    np.savetxt('%s'%outfile, data, fmt='%s', delimiter=delimiter)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
import numpy as np
import netCDF4 as nc
import scipy.io as sio
import os
from collections import OrderedDict
from pyjams.jams.writenetcdf import writenetcdf


def mat2nc(fname, overwrite=False, fname_out=None, verbose=True, squeeze=True,
           varname=None,
           lat=None, lon=None, lat_file=None, lon_file=None, transposed=False,
           regular_grid=True,
           anchor_time=None, data_time=None):
    """
        This functions writes the content of a Matlab file into a NetCDF file.
        It allows for specifying longitudes and latitudes as well as a validity time.

        Definition
        ----------
        def mat2nc( fname )


        Input           Format                  Description
        -----           -----                   -----------
        fname           (array-like of) string  Filename of matlab file.
                                                If a list of file names is given, multiple files will be read and written in one netcdf file.
                                                This feature is implemented such that multiple time steps can be written to one file.
                                                Make sure that all files specified have the same structure (e.g. same grid) and that only
                                                one variable is stored per file. The "varname", "fname_out", "anchor_time" and an array of
                                                "data_time" should be specified in this case.
        overwrite       bool                    Optional argument.
                                                True:  existing nc-file will be overwritten,
                                                False: if nc-file exists, script will be stopped
                                                Default: False
        fname_out       string                  Optional argument.
                                                Name of output NetCDF file
                                                Default: abc.mat --> abc.nc
        verbose         bool                    Optional argument.
                                                True:  variable names found etc. are print on screen
                                                False: silent
                                                Default: True
        squeeze         bool                    Optional argument.
                                                True:  all axis with shape 1 are removed
                                                False: axis with shape 1 are written (arrays will have same shape like in Matlab)
                                                Default: True
        varname         (array-like of) string  Optional argument.
                                                Use this name for variable in NetCDF file.
                                                If there are multiple variables in one Matlab file a list of
                                                varnames should be specified.
                                                Default: Name of variable in Matlab file
        lat             array-like              Optional argument for input specifying latitude
                                                (needs to match dimensions of all variables read from mat-file)
        lon             array-like              Optional argument for input specifying longitude
                                                (needs to match dimensions of all variables read from mat-file)
        lat_file        string                  Optional argument for input file (*.mat) specifying latitude
        lon_file        string                  Optional argument for input file (*.mat) specifying longitude
        transposed      boolean                 Optional argument. If true all gridded data will be transposed.
                                                This might be necessary if data are stored with swapped lon and lat,
                                                i.e. longitude is y-axis and latitude is x-axis.
        anchor_time     string                  Time where data in Matlab file are valid. Has to follow the pattern
                                                    'hours since YYYY-MM-DD HH:MM:SS'
        data_time       array-like              Time of each data set written, i.e. if there are 4 precipitations fields and
                                                they are valid 6, 12, 18, 24 hrs after anchor time, data_time=[6,12,18,24].


        Description
        -----------
        None

        Restrictions
        ------------
        None

        Examples
        --------
        >>> import scipy.io as sio
        >>> import os, jams
        >>> ss = np.array([[10.0]])
        >>> tt = np.array([[10.0, 1.0, 1.0],[20.0, 2.0, 2.0]])
        >>> uu = np.array([[[10.0], [1.0], [1.0]],[[20.0], [2.0], [2.0]]])
        >>> vv = np.array([10.0,1.0, 1.0, 20.0, 2.0, 2.0])
        >>> sio.savemat('test.mat', {'ss':ss, 'tt':tt, 'uu':uu, 'vv':vv})

        # --------------------------------------------
        # Write squeezed arrays (all axis with shape 1 are removed)
        # --------------------------------------------
        >>> if os.path.isfile('test.nc'): os.remove('test.nc')
        >>> mat2nc('test.mat')
        writes  'ss'                  shape =  (1,)                  to *.nc
        writes  'tt'                  shape =  (2, 3)                to *.nc
        writes  'uu'                  shape =  (2, 3)                to *.nc
        writes  'vv'                  shape =  (6,)                  to *.nc
        wrote file  test.nc

        >>> from readnetcdf import readnetcdf
        >>> print(readnetcdf('test.nc',var='ss'))
        [10.]
        >>> print(readnetcdf('test.nc',var='tt'))
        [[10.  1.  1.]
         [20.  2.  2.]]
        >>> print(readnetcdf('test.nc',var='uu'))
        [[10.  1.  1.]
         [20.  2.  2.]]
        >>> print(readnetcdf('test.nc',var='vv'))
        [10.  1.  1. 20.  2.  2.]

        # --------------------------------------------
        # Write un-squeezed, original arrays (same shape as in Matlab)
        # --------------------------------------------
        >>> if os.path.isfile('test.nc'): os.remove('test.nc')
        >>> mat2nc('test.mat',squeeze=False,verbose=False)

        >>> print(readnetcdf('test.nc',var='ss'))
        [[10.]]
        >>> print(readnetcdf('test.nc',var='tt'))
        [[10.  1.  1.]
         [20.  2.  2.]]
        >>> print(readnetcdf('test.nc',var='uu'))
        [[[10.]
          [ 1.]
          [ 1.]]
         [[20.]
          [ 2.]
          [ 2.]]]
        >>> print(readnetcdf('test.nc',var='vv'))
        [[10.  1.  1. 20.  2.  2.]]

        >>> if os.path.isfile('test.nc'): os.remove('test.nc')
        >>> if os.path.isfile('test.mat'): os.remove('test.mat')


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2016 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Sep 2016
        Modified, JM, Oct 2016 - add feature xy
                  MC, Nov 2016 - single file not working in Python 3, Order .mat content by name (for docstrings).
    """

    # make sure that fname is an array
    import sys
    if sys.version_info > (3,0): basestring = str
    if hasattr(fname, "__iter__") and not isinstance(fname, basestring):
        filename_mat = np.array(fname)
    else:
        filename_mat = np.array([fname])

    # make sure that varname is an array
    if varname is not None:
        if hasattr(varname, "__iter__"):
            varname = np.array(varname)
        else:
            varname = np.array([varname])

    # make sure that data_time is an array
    if data_time is not None:
        if hasattr(data_time, "__iter__"):
            data_time = np.array(data_time)
        else:
            data_time = np.array([data_time])

    # make sure that output filename is a valid one
    if fname_out is None:
        filename_nc  = '.'.join(filename_mat[0].split('.')[:-1])+'.nc'
        if filename_nc == '.nc': # in case input filename is missing file ending ".mat"
            filename_nc = 'new.nc'
    else:
        filename_nc  = fname_out

    # make sure that output file does not exist
    if os.path.isfile(filename_nc) :
        if overwrite:
            # remove existing nc-file
            os.remove(filename_nc)
        else:
            # stop script
            raise ValueError('mat2nc: NetCDF file already exists')

    # assure that time points are sorted
    if ( not (data_time is None) ):
        idx = np.argsort(data_time)
        data_time = data_time[idx]
        filename_mat = filename_mat[idx]

    # checking of input arguments
    if (not(lon is None) and lat is None) or (lon is None and not(lat is None)):
        raise ValueError('mat2nc: either both lat and lon have to be specified or none of it')

    if (not(lon_file is None) and lat_file is None) or (lon_file is None and not(lat_file is None)):
        raise ValueError('mat2nc: either both lat_file and lon_file have to be specified or none of it')

    if (not(lat_file is None) and not(lat is None)):
        raise ValueError('mat2nc: grid information (lat/lon) can either be given as file or as array but not both')

    # initialize counters and variables
    key_content  = None
    grid_avail   = (not (lon is None)) or (not (lon_file is None))
    time_avail   = not (anchor_time is None)
    dim_name     = np.array([ 'dim_'+str(ii) for ii in range(100) ])  # ['dim_0', 'dim_1', 'dim_2', ...., 'dim_99']
    dims         = np.zeros(np.shape(dim_name),dtype=int)
    idim         = 0 # index where to write next dim

    # Read longitude info from file
    cc = 0
    if (not lon_file is None):
        if not os.path.isfile(lon_file):
            raise ValueError('mat2nc: Longitude file not found')
        lon_contents = sio.loadmat(lon_file)
        for ikey,key in enumerate(lon_contents):
            if key[0:2] != '__':
                cc += 1
                lon = np.array(lon_contents[key])

                if transposed:
                    lon = np.transpose(lon)

        if regular_grid:
            # look if longitudes are constant per column
            lon_tmp   = np.round(lon,3)
            lon_const = np.all([ list(lon_tmp[:,ii]).count(lon_tmp[0,ii]) == len(lon_tmp[:,ii]) for ii in range(np.shape(lon_tmp)[1])])

            if not lon_const:
                print('lon[:,0]: ',lon_tmp[:,0])
                raise ValueError('mat2nc: Longitudes are not constant per column. Use either irregular=True or try transpose=True')

    if (cc > 1):
        raise ValueError('mat2nc: More than one variable found in longitude file')


    # Read latitude info from file
    cc = 0
    if (not lat_file is None):
        if not os.path.isfile(lat_file):
            raise ValueError('mat2nc: Latitude file not found')
        lat_contents = sio.loadmat(lat_file)
        for ikey,key in enumerate(lat_contents):
            if key[0:2] != '__':
                cc += 1
                lat = np.array(lat_contents[key])

                if transposed:
                    lat = np.transpose(lat)

        if regular_grid:
            # look if longitudes are constant per column
            lat_tmp   = np.round(lat,3)
            lat_const = np.all([ list(lat_tmp[ii,:]).count(lat_tmp[ii,0]) == len(lat_tmp[ii,:]) for ii in range(np.shape(lat_tmp)[0])])

            if not lat_const:
                print('lat[0,:]: ',lat_tmp[0,:])
                raise ValueError('mat2nc: Latitudes are not constant per row. Use either irregular=True or try transpose=True')

    if (cc > 1):
        raise ValueError('mat2nc: More than one variable found in latitude file')

    # open netcdf file and add some general information
    fh = nc.Dataset( filename_nc, 'w', 'NETCDF4' )
    # File attributes
    FiAtt   = ([['description', 'Dump content of '+", ".join(filename_mat)+' to NetCDF'],
                ['history'    , 'Created by Juliane Mai (mat2nc.py)']])
    handle  = writenetcdf(fh, fileattributes=FiAtt)

    dd = 0
    # if time available, write this dimensions
    if (time_avail):
        varName     = 'ntime'
        varAtt  = ([['units',    anchor_time],
                    ['calendar', 'gregorian']])
        ntime        = np.shape(np.array(data_time))[0]
        dims[dd]     = 1 # means that only one time step is written
        dim_name[dd] = varName
        # create dimension
        th = writenetcdf(fh, name=dim_name[dd], dims=None, attributes=varAtt, isdim=True)
        # write values to dimension
        times   = np.array(data_time)
        writenetcdf(fh, th, time=range(ntime), var=times)

        dd += 1

    # if grid available, write this dimensions and data first
    if (grid_avail):

        # (1) check if variable shape matches given grid shape
        arr_shape_1 = np.shape(lat)
        arr_shape_2 = np.shape(lon)
        if (arr_shape_1 != arr_shape_2):
            raise ValueError('mat2nc: shapes of given lat and lon grids are not matching')

        # TODO: If dim_lat = dim_lon the identification of lon and lat axis in variables might not work properly
        if (np.shape(lon)[0] == np.shape(lat)[1]):
            print('WARNING: ')
            raise ValueError('mat2nc: dimension of lat and lon are equal. This is not supported yet.')

        # (2) write dimensions and lat/lon data
        #        lon is x-axis ==> number columns
        #        lat is y-axis ==> number of rows
        varName     = 'nlon'
        varAtt      = ([['units'         , 'degrees_east'],
                        ['standard_name' , 'longitude'   ],
                        ['missing_value' , -9.]])
        nlon         = np.shape(lon)[1]
        dims[dd]     = nlon
        dim_name[dd] = varName
        if regular_grid:
            # create dimension and write variable together
            var          = lon[0] #lon[:,0]
            writenetcdf(fh, name=dim_name[dd], dims=dims[dd], var=var, attributes=varAtt, isdim=True)
        else:
            # just create dimension but don't write variable yet
            writenetcdf(fh, name=dim_name[dd], dims=dims[dd],
                        var=None, isdim=True, create_var=False )
        dd += 1
        varName     = 'nlat'
        varAtt      = ([['units'         , 'degrees_north'],
                        ['standard_name' , 'latitude'     ],
                        ['missing_value' , -9.]])
        nlat         = np.shape(lat)[0]
        dims[dd]     = nlat
        dim_name[dd] = varName
        if regular_grid:
            var          = lat[:,0] #lat[0]
            writenetcdf(fh, name=dim_name[dd], dims=dims[dd], var=var, attributes=varAtt, isdim=True)
        else:
            # just create dimension but don't write variable yet
            writenetcdf(fh, name=dim_name[dd], dims=dims[dd],
                        var=None, isdim=True, create_var=False )
        dd += 1

        # now both dimensions exist and grid can be written as extra variable
        if not regular_grid:
            idx=[ np.where(dims==dd)[0][0] for dd in np.shape(lon) ]

            attributes = {'standard_name': 'longitude', 'long_name': 'longitudes of irregular grid', 'units': 'degrees_??', 'missing_value': -9.}
            writenetcdf(fh, name='lon', dims=dim_name[idx], var=lon,
                        comp=True, create_var=True,
                        attributes=attributes,
                        vartype=lon.dtype )

            idx=[ np.where(dims==dd)[0][0] for dd in np.shape(lat) ]
            attributes = {'standard_name': 'latitude', 'long_name': 'latitudes of irregular grid', 'units': 'degrees_??', 'missing_value': -9.}
            writenetcdf(fh, name='lat', dims=dim_name[idx], var=lat,
                        comp=True, create_var=True,
                        attributes=attributes,
                        vartype=lat.dtype )


    tt = 0
    for ff in filename_mat:
        vv = 0

        # read content from Matlab file
        mat_contents = sio.loadmat(ff)
        mat_contents = OrderedDict(sorted(mat_contents.items(), key=lambda t: t[0]))

        for ikey, key in enumerate(mat_contents):
            if key[0:2] != '__':

                key_content = np.array(mat_contents[key])
                tmp = key_content

                if transposed:
                    tmp = np.transpose(tmp)

                arr_shape = np.shape(tmp)

                # if there is exactly one entry for time_data given, i.e. the data read is the data for exactly one time point
                # but it might be that this dimension is missing, i.e. data_shape=[lat, lon] but should be data_shape=[1, lat, lon]
                if (time_avail):
                    if (not (1 in arr_shape)):
                        tmp = np.expand_dims(tmp,axis=0)
                        arr_shape = np.shape(tmp)

                if (not grid_avail and not time_avail):
                    # remove axis with shape 1, i.e.
                    #        shape=[1,23,30,1,4] --> shape=[23,30,4]
                    # but    shape=[1,1,1]       --> shape=[1] (otherwise cannot be written to NetCDF)
                    if (squeeze):
                        if  (hasattr(tmp, "__len__")):
                            tmp = np.squeeze(tmp)

                            # in case tmp contains only one scalar, make again proper array of it
                            if (tmp.shape == ()):
                                tmp = np.array([tmp])

                if (not (varname is None)):
                    key = varname[vv]

                if verbose:
                    print('writes ',repr(key).ljust(35),' shape = ',repr(np.shape(tmp)).ljust(20),' to *.nc')

                # write (missing) dimensions
                arr_shape = np.shape(tmp)
                for dd in range(len(arr_shape)):
                    if (not arr_shape[dd] in dims) and (not dim_name[idim] in ['nlat','nlon','ntime']):

                        dims[idim] = arr_shape[dd]
                        writenetcdf(fh, name=dim_name[idim], dims=dims[idim],
                                    var=None, isdim=True, create_var=False )
                        idim +=1
                        if (idim > len( dim_name )):
                            raise ValueError( 'mat2nc: Too many different dimensions found. You may have to increase size of dim_name.' )

                # find matching dim names for current shape
                idx=[ np.where(dims==arr_shape[dd])[0][0] for dd in range(len(arr_shape)) ]

                if (grid_avail):
                    # check if dimension matching lon and lat are found for this variable
                    if (not 'nlat' in dim_name[idx]) or (not 'nlon' in dim_name[idx]):
                        if verbose:
                            print('    WARNING: lat and/or lon dimension not found for variable')

                attributes = {'long_name': key, 'units': 'not defined'}
                if (not (varname is None)):
                    if (varname[vv].lower() == 'av'):
                        attributes = {'long_name': 'Accumulated surface flux of latent heat (accumulated FV)', 'units': 'J/m^2'}
                    elif (varname[vv].lower() == 'fb'):
                        attributes = {'long_name': 'Downward solar flux', 'units': 'W/m^2'}
                    elif (varname[vv].lower() == 'fi'):
                        attributes = {'long_name': 'Surface incoming infrared flux', 'units': 'W/m^2'}
                    elif (varname[vv].lower() == 'hu'):
                        attributes = {'long_name': 'Specific humidity', 'units': 'kg/kg'}
                    elif (varname[vv].lower() == 'p0'):
                        attributes = {'long_name': 'Surface pressure', 'units': 'mbar'}
                    elif (varname[vv].lower() == 'pr'):
                        attributes = {'long_name': 'Precipitation', 'units': 'mm'}
                    elif (varname[vv].lower() == 'tt'):
                        attributes = {'long_name': 'Air temperature', 'units': 'degree C'}
                    elif (varname[vv].lower() == 'uu'):
                        attributes = {'long_name': 'U-component of the wind (along the grid X axis)', 'units': 'kts'}
                    elif (varname[vv].lower() == 'vv'):
                        attributes = {'long_name': 'V-component of the wind (along the grid Y axis)', 'units': 'kts'}

                if (time_avail):
                    if (tt==0):
                        vh   = writenetcdf(fh, name=key, dims=dim_name[idx],
                                           comp=True,
                                           attributes=attributes,
                                           vartype=key_content.dtype)
                    writenetcdf(fh, vh, var = tmp[0], time = tt)
                    tt += 1
                else:
                    writenetcdf(fh, name=key, dims=dim_name[idx], var=tmp,
                                comp=True,
                                attributes=attributes,
                                vartype=key_content.dtype)
                vv += 1

    fh.close()

    if verbose:
        print('wrote file ',filename_nc)

if __name__ == '__main__':
    import doctest
    try:
        import netCDF4 as nc
        doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    except:
        raise IOError('No NetCDF4 support available.')



#!/usr/bin/env python
import numpy as np


__all__ = ['constant_values']


# --------------------------------------------------------------------

def constant_values(juldate,series, length, eps):
    """
        Checks if a given series of data contains consecutive values which are constant over a certain time period.

        Definition
        ----------
        def constant_values(juldate, series, length, eps):


        Input
        -----
        juldate 1D-array of julian dates for data points in series
        series  1D-array of data
        length  length of time period in which values are checked for constancy
                (unit of length is days)
        eps     difference of minimal and maximal data point in window of length <length> has to
                be above 2*eps otherwise all datapoints in window are regarded to be constant


        Output
        ------
        1D-array of boolean values indicating if value is in a constant period (.true.) or not (.false.)


        Examples
        --------
        --> see __init__.py for full example of workflow

        >>> dates  = np.array([100.1, 100.6, 101.1, 101.6, 102.1, 102.6, 103.1, 103.6, 104.1, 104.6, 105.1, 105.6])
        >>> series = np.array([1.0, 2.0, 5.0, 3.000, 3.001, 3.010, 3.012, 2.990, 3.100, 4.0, 3.0, 5.0])
        >>> length = 2
        >>> eps    = 0.1
        >>> print(constant_values(dates, series, length, eps))
        [False False False  True  True  True  True  True  True False False False]

        >>> dates  = np.array([100.1, 100.6, 101.1, 101.6, 102.1, 102.6, 103.1, 103.6, 104.1, 104.6, 105.1, 105.6])
        >>> series = np.array([1.0, 2.0, 5.0, 3.000, 3.001, 3.010, 3.012, 2.990, 3.100, 4.0, 3.0, 5.0])
        >>> length = 3.5
        >>> eps    = 0.1
        >>> print(constant_values(dates, series, length, eps))
        [False False False False False False False False False False False False]

        >>> dates  = np.array([100.1, 100.6, 101.1, 101.6, 102.1, 102.6, 103.1, 103.6, 104.1, 104.6, 105.1, 105.6])
        >>> series = np.array([1.0, 2.0, 5.0, 3.000, 3.001, 4.0, 3.012, 2.990, 3.100, 4.0, 3.0, 5.0])
        >>> length = 2
        >>> eps    = 0.1
        >>> print(constant_values(dates, series, length, eps))
        [False False False False False False False False False False False False]

        >>> dates  = np.array([100.1, 100.6, 101.1, 101.6, 102.1, 102.6, 103.1, 103.6, 104.1, 104.6, 105.1, 105.6])
        >>> series = np.array([1.0, 2.0, 5.0, 4.0, 3.0, 5.0, 4.0, 3.000, 3.001, 3.012, 2.990, 3.100])
        >>> length = 2
        >>> eps    = 0.1
        >>> print(constant_values(dates, series, length, eps))
        [False False False False False False False  True  True  True  True  True]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Juliane Mai

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Oct 2015
    """
    if (len(series.shape) > 1):
        raise ValueError('constant_values: only allowed for 1d data arrays')

    if (len(juldate.shape) > 1):
        raise ValueError('constant_values: only allowed for 1d julian date arrays')

    if (np.size(juldate,0) != np.size(series,0) ):
        raise ValueError('constant_values: size of dates and data points are not matching')

    nseries = np.size(series,0)

    out = np.zeros(nseries, dtype=bool)

    # indexes of entries where difference is too small
    diff_idx = np.where(np.abs(np.diff(series))<2.0*eps)[0]

    for idx in diff_idx:

        wind_idx = np.where((juldate>=juldate[idx]) & (juldate<juldate[idx]+length))[0]

        min_val = np.amin(series[wind_idx])
        max_val = np.amax(series[wind_idx])

        while (max_val - min_val < 2.0*eps):

            out[wind_idx] = True

            # try now also with next data point
            # but only if last index is not the last one possible
            if (wind_idx[-1] != nseries - 1):
                wind_idx = np.append(wind_idx,np.array(wind_idx[-1]+1))
                min_val = np.amin(series[wind_idx])
                max_val = np.amax(series[wind_idx])
            else:
                # make sure that it exits the while here
                max_val = min_val + 3.0*eps

        # shorten the idexes which still have to be checked
        # e.g. [2,3,4,5,6,10,11,12] --> [10,11,12]
        diff_idx = diff_idx[np.where(diff_idx>wind_idx[-1]-1)[0]]

    return out

# --------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
    File list functions.


    Provided functions
    ------------------
    fullnames                   Filenames with absolute paths in local directories.
    fullnames_dates             Filenames with absolute paths and modification times in local directories.
    fullnames_dates_sizes       Filenames with absolute paths, modification times, and file sizes in local directories.
    fullnames_sizes             Filenames with absolute paths and file sizes in local directories.
    fullnames_times             Wrapper for fullnames_dates.
    fullnames_times_sizes       Wrapper for fullnames_dates_sizes.
    last_fullname               Filename with absolute paths of last file in local directories.
    last_fullname_date          Filename with absolute paths and modification time of last file in local directories.
    last_fullname_date_size     Filename with absolute paths, modification time, and
                                file size of last file in local directories.
    last_fullname_size          Filename with absolute paths and file size of last file in local directories.
    last_fullname_time          Wrapper for last_fullname_date.
    last_fullname_time_size     Wrapper for last_fullname_date_size.
    last_name                   Filename of last file in local directories.
    last_name_date              Filename and modification time of last file in local directories.
    last_name_date_size         Filename, modification time, and file size of last file in local directories.
    last_name_size              Filename and file size of last file in local directories.
    last_name_time              Wrapper for last_name_date.
    last_name_date_size         Wrapper for last_name_date_size.
    names                       Filenames in local directories.
    names_dates                 Filenames and modification times in local directories.
    names_dates_sizes           Filenames, modification times, and file sizes in local directories.
    names_sizes                 Filenames and file sizes in local directories.
    names_times                 Wrapper for names_dates.
    names_times_sizes           Wrapper for names_dates_sizes.
    newest_fullname             Wrapper for last_fullname.
    newest_fullname_date        Wrapper for last_fullname_date.
    newest_fullname_date_size   Wrapper for last_fullname_date_size.
    newest_fullname_size        Wrapper for last_fullname_size.
    newest_fullname_time        Wrapper for last_fullname_date.
    newest_fullname_time_size   Wrapper for last_fullname_date_size.
    newest_name                 Wrapper for last_name.
    newest_name_date            Wrapper for last_name_date.
    newest_name_date_size       Wrapper for last_name_date_size.
    newest_name_size            Wrapper for last_name_size.
    newest_name_time            Wrapper for newest_name_date.
    newest_name_date_size       Wrapper for newest_name_date_size.


    Example
    -------
    import os
    # get .dat filenames with asolute paths in directories 2013, 2014, ...
    fls = fullnames('*.dat', dirs=glob.glob('[0-9][0-9][0-9][0-9]'))

    # get .dat filenames and sizes in current directory
    os.chdir('change/directory')
    fls, flssize = names_sizes('*.dat')

    # get all filenames and modification times in current directory
    fls, flstimes = names_times()


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC, Jun-Dec 2014
"""
from .files import *

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.0'
__revision__ = "Revision: 1923"
__date__     = 'Date: 05.12.2014'



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
lowess: Locally linear regression  
=================================

Implementation of the LOWESS algorithm in n dimensions.


References
==========
[HTF08] Hastie, Tibshirani and Friedman (2008).
The Elements of Statistical Learning; Chapter 6.

[Cleveland79] Cleveland (1979).
Robust Locally Weighted Regression and Smoothing Scatterplots.
J American Statistical Association, 74: 829-836.


Author
==========
Ariel Rokem, arokem@gmail.com, http://github.com/arokem/lowess
Revision 0.1_dev

Copyright (c) 2015--, Ariel Rokem
All rights reserved.


License
=======
Copyright (c) 2015 Ariel Rokem

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import numpy.linalg as la

# __all__ = ['epanechnikov', 'tri_cube', 'bi_square', 'lowess']
__all__ = ['lowess']

# Kernel functions
def epanechnikov(xx, **kwargs):
    """
    The Epanechnikov kernel estimated for xx values at indices idx (zero
    elsewhere) 

    Parameters
    ----------
    xx: float array
        Values of the function on which the kernel is computed. Typically,
        these are Euclidean distances from some point x0 (see do_kernel)

    Notes
    -----
    This is equation 6.4 in HTF08 chapter 6        
    
    """
    l = kwargs.get('l', 1.0)
    ans = np.zeros(xx.shape)
    xx_norm = xx / l
    idx = np.where(xx_norm <= 1)
    ans[idx] = 0.75 * (1 - xx_norm[idx]**2)
    return ans


def tri_cube(xx, **kwargs):
    """ 
    The tri-cube kernel estimated for xx values at indices idx (zero
    elsewhere) 

    Parameters
    ----------
    xx: float array
        Values of the function on which the kernel is computed. Typically,
        these are Euclidean distances from some point x0 (see do_kernel)

    idx: tuple
        An indexing tuple pointing to the coordinates in xx for which the
        kernel value is estimated. Default: None (all points are used!)  

    Notes
    -----
    This is equation 6.6 in HTF08 chapter 6        
    """        
    ans = np.zeros(xx.shape)
    idx = np.where(xx <= 1)
    ans[idx] = (1 - np.abs(xx[idx])**3)**3
    return ans


def bi_square(xx, **kwargs):
    """
    The bi-square weight function calculated over values of xx

    Parameters
    ----------
    xx: float array

    Notes
    -----
    This is the first equation on page 831 of [Cleveland79].
    """
    ans = np.zeros(xx.shape)
    idx = np.where(xx < 1)
    ans[idx] = (1 - xx[idx]**2)**2
    return ans


# Apply kernel
def do_kernel(x0, x, l=1.0, kernel=epanechnikov):
    """
    Calculate a kernel function on x in the neighborhood of x0

    Parameters
    ----------
    x: float array
       All values of x
    x0: float
       The value of x around which we evaluate the kernel
    l: float or float array (with shape = x.shape)
       Width parameter (metric window size)
    kernel: callable
        A kernel function. Any function with signature: `func(xx)`    
    """
    # xx is the norm of x-x0. Note that we broadcast on the second axis for the
    # nd case and then sum on the first to get the norm in each value of x:
    xx = np.sum(np.sqrt(np.power(x - x0[:, np.newaxis], 2)), 0)
    return kernel(xx, l=l)


# Lowess
def lowess(x, y, x0, deg=1, kernel=epanechnikov, l=1, robust=False):
    """
    Locally smoothed regression with the LOWESS algorithm.

    Parameters
    ----------
    x: float n-d array  
       Values of x for which f(x) is known (e.g. measured). The shape of this
       is (n, j), where n is the number the dimensions and j is the
       number of distinct coordinates sampled.
    
    y: float array
       The known values of f(x) at these points. This has shape (j,) 

    x0: float or float array.
        Values of x for which we estimate the value of f(x). This is either a
        single scalar value (only possible for the 1d case, in which case f(x0)
        is estimated for just that one value of x), or an array of shape (n, k).

    deg: int
        The degree of smoothing functions. 0 is locally constant, 1 is locally
        linear, etc. Default: 1.
        
    kernel: callable
        A kernel function. {'epanechnikov', 'tri_cube', 'bi_square'}

    l: float or float array with shape = x.shape
       The metric window size for the kernel

    robust: bool
        Whether to apply the robustification procedure from [Cleveland79], page
        831
    
        
    Returns
    -------
    The function estimated at x0. 

    Notes
    -----
    The solution to this problem is given by equation 6.8 in Hastie
    Tibshirani and Friedman (2008). The Elements of Statistical Learning
    (Chapter 6). 

    Example
    -------
    >>> import lowess as lo
    >>> import numpy as np

    # For the 1D case:
    >>> x = np.random.randn(100)
    >>> f = np.cos(x) + 0.2 * np.random.randn(100)
    >>> x0 = np.linspace(-1, 1, 10)
    >>> f_hat = lo.lowess(x, f, x0)
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots(1)
    >>> _ = ax.scatter(x, f)
    >>> _ = ax.plot(x0, f_hat, 'ro')
    >>> plt.show()
    >>> None

    # 2D case (and more...)
    >>> x = np.random.randn(2, 100)
    >>> f = -1 * np.sin(x[0]) + 0.5 * np.cos(x[1]) + 0.2*np.random.randn(100)
    >>> x0 = np.mgrid[-1:1:.1, -1:1:.1]
    >>> x0 = np.vstack([x0[0].ravel(), x0[1].ravel()])
    >>> f_hat = lo.lowess(x, f, x0, kernel=lo.tri_cube)
    >>> from mpl_toolkits.mplot3d import Axes3D
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> _ = ax.scatter(x[0], x[1], f)
    >>> _ = ax.scatter(x0[0], x0[1], f_hat, color='r')
    >>> plt.show()
    >>> None
    """
    if robust:
        # We use the procedure described in Cleveland1979
        # Start by calling this function with robust set to false and the x0
        # input being equal to the x input:
        y_est = lowess(x, y, x, kernel=epanechnikov, l=l, robust=False)
        resid = y_est - y
        median_resid = np.nanmedian(np.abs(resid))
        # Calculate the bi-cube function on the residuals for robustness
        # weights: 
        robustness_weights = bi_square(resid / (6 * median_resid))
        
    # For the case where x0 is provided as a scalar: 
    if not np.iterable(x0):
       x0 = np.asarray([x0])
    ans = np.zeros(x0.shape[-1]) 
    # We only need one design matrix for fitting:
    B = [np.ones(x.shape[-1])]
    for d in range(1, deg+1):
        B.append(x**deg)

    B = np.vstack(B).T
    for idx, this_x0 in enumerate(x0.T):
        # This is necessary in the 1d case (?):
        if not np.iterable(this_x0):
            this_x0 = np.asarray([this_x0])
        # Different weighting kernel for each x0:
        W = np.diag(do_kernel(this_x0, x, l=l, kernel=kernel))
        # XXX It should be possible to calculate W outside the loop, if x0 and
        # x are both sampled in some regular fashion (that is, if W is the same
        # matrix in each iteration). That should save time.

        if robust:
            # We apply the robustness weights to the weighted least-squares
            # procedure:
            robustness_weights[np.isnan(robustness_weights)] = 0
            W = np.dot(W, np.diag(robustness_weights))
        #try: 
        # Equation 6.8 in HTF08:
        # BtWB = np.dot(np.dot(B.T, W), B)
        # BtW = np.dot(B.T, W)
        BtW = np.dot(B.T, W)
        BtWB = np.dot(BtW, B)
        # Get the params:
        beta = np.dot(np.dot(la.pinv(BtWB), BtW), y.T)
        # We create a design matrix for this coordinat for back-predicting:
        B0 = [1]
        for d in range(1, deg+1):
            B0 = np.hstack([B0, this_x0**deg])
        B0 = np.vstack(B0).T
        # Estimate the answer based on the parameters:
        ans[idx] += np.dot(B0, beta)
    # If we are trying to sample far away from where the function is
    # defined, we will be trying to invert a singular matrix. In that case,
    # the regression should not work for you and you should get a nan:
        #except la.LinAlgError :
        #    ans[idx] += np.nan
    return ans.T


if __name__ == '__main__':
    # import doctest
    # doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    np.random.seed(1234)
    n = 101
    x = np.linspace(0, 6, n)
    y = np.sin(x) + 0.3 * np.random.randn(n)
    xout = np.linspace(0, 6-1, 30) + 0.3

    f = 0.25
    # yest = lowess(x, y, x, l=f)
    yout = lowess(x, y, xout, l=f)

    # import pylab as plt
    # plt.clf()
    # plt.plot(x, y, 'bo', label='y noisy')
    # plt.plot(x, yest, label='y pred')
    # plt.plot(xout, yout, 'ro', label='y pred xout')
    # plt.legend()
    # plt.show()
